from __future__ import annotations

from llmling.resources import (
    ResourceLoader,
    LoadedResource,
    default_registry as resource_registry,
)
from llmling.core.exceptions import (
    LLMLingError,
    ConfigError,
    ResourceError,
    LoaderError,
    ProcessorError,
    LLMError,
)
from llmling.processors.registry import ProcessorRegistry
from llmling.server import LLMLingServer, serve


__version__ = "0.6.4"

__all__ = [
    "ConfigError",
    "LLMError",
    "LLMLingError",
    "LLMLingServer",
    "LoadedResource",
    "LoaderError",
    "ProcessorError",
    "ProcessorRegistry",
    "ResourceError",
    "ResourceLoader",
    "resource_registry",
    "serve",
]

# llmling/
# ├── src/
# │   └── llmling/
# │       ├── __init__.py                 # Main package exports
# │       ├── client.py                   # High-level client interface
# │       │
# │       ├── core/                       # Core components
# │       │   ├── __init__.py
# │       │   ├── capabilities.py         # LLM model capabilities
# │       │   ├── descriptors.py          # Python descriptors
# │       │   ├── exceptions.py           # Exception hierarchy
# │       │   ├── log.py                  # Logging configuration
# │       │   ├── typedefs.py            # Common type definitions
# │       │   ├── utils.py               # Generic utilities
# │       │   └── baseregistry.py        # Base registry class
# │       │
# │       ├── config/                     # Configuration handling
# │       │   ├── __init__.py
# │       │   ├── models.py              # Configuration models
# │       │   ├── loading.py             # Config loading utilities
# │       │   ├── manager.py             # Config management
# │       │   └── validation.py          # Config validation
# │       │
# │       ├── resources/                    # Resource handling
# │       │   ├── __init__.py
# │       │   ├── base.py                # Base resource classes
# │       │   ├── models.py              # Resource models
# │       │   ├── registry.py            # Resource registry
# │       │   └── loaders/               # Resource loaders
# │       │       ├── __init__.py
# │       │       ├── callable.py
# │       │       ├── cli.py
# │       │       ├── image.py
# │       │       ├── path.py
# │       │       ├── source.py
# │       │       └── text.py
# │       │
# │       ├── processors/                 # Resource processing
# │       │   ├── __init__.py
# │       │   ├── base.py                # Base processor classes
# │       │   ├── registry.py            # Processor registry
# │       │   └── implementations/       # Processor implementations
# │       │       ├── __init__.py
# │       │       ├── function.py
# │       │       └── template.py
# │       │
# │       ├── tools/                     # Tool system
# │       │   ├── __init__.py
# │       │   ├── base.py                # Base tool classes
# │       │   ├── actions.py             # Action definitions
# │       │   ├── browser.py             # Browser automation
# │       │   ├── code.py                # Code analysis
# │       │   └── exceptions.py          # Tool exceptions
# │       │
# │       ├── utils/                     # Utilities
# │       │   ├── __init__.py
# │       │   ├── importing.py           # Import utilities
# │       │   └── calling.py             # Callable utilities
# │       │
# │       ├── testing/                   # Testing utilities
# │       │   ├── __init__.py
# │       │   ├── processors.py          # Test processors
# │       │   └── tools.py               # Test tools
# │       │
# │       └── config_resources/                 # Configuration resources
# │           ├── test.yml              # Test configuration
# │           └── web_research.yml      # Web research config
# │
# ├── tests/                             # Test suite
# ├── examples/                          # Example usage
# ├── docs/                             # Documentation
# └── resources/                        # Resource files



"""MCP server implementation."""

from __future__ import annotations

import asyncio
import logging
from typing import TYPE_CHECKING, Any, Self

import mcp
from mcp.server import Server
from mcp.types import (
    INTERNAL_ERROR,
    INVALID_PARAMS,
    CompleteResult,
    Completion,
    CompletionArgument,
    GetPromptResult,
    TextContent,
)
from pydantic import AnyUrl

from llmling import config_resources
from llmling.config.models import PathResource, SourceResource
from llmling.core.log import get_logger
from llmling.processors.registry import ProcessorRegistry
from llmling.prompts.models import ArgumentType, Prompt
from llmling.prompts.registry import PromptRegistry
from llmling.resources import ResourceLoaderRegistry
from llmling.resources.registry import ResourceRegistry
from llmling.server import conversions
from llmling.server.log import configure_server_logging, run_logging_processor
from llmling.server.observers import PromptObserver, ResourceObserver, ToolObserver
from llmling.tools.registry import ToolRegistry


if TYPE_CHECKING:
    from collections.abc import Coroutine
    import os

    from llmling.config.models import Config

logger = get_logger(__name__)

DEFAULT_NAME = "llmling-server"


class LLMLingServer:
    def __init__(
        self,
        config: Config,
        *,
        name: str = DEFAULT_NAME,
        loader_registry: ResourceLoaderRegistry | None = None,
        processor_registry: ProcessorRegistry | None = None,
        prompt_registry: PromptRegistry | None = None,
        tool_registry: ToolRegistry | None = None,
    ) -> None:
        """Initialize server."""
        self.config = config
        self.name = name

        # Initialize registries
        self.loader_registry = loader_registry or ResourceLoaderRegistry()
        self.processor_registry = processor_registry or ProcessorRegistry()
        self.prompt_registry = prompt_registry or PromptRegistry()
        self.tool_registry = tool_registry or ToolRegistry()

        # Create resource registry with dependencies
        self.resource_registry = ResourceRegistry(
            loader_registry=self.loader_registry,
            processor_registry=self.processor_registry,
        )

        # Register default resource loaders if using new registry
        if loader_registry is None:
            from llmling.resources import (
                CallableResourceLoader,
                CLIResourceLoader,
                ImageResourceLoader,
                PathResourceLoader,
                SourceResourceLoader,
                TextResourceLoader,
            )

            self.loader_registry["text"] = TextResourceLoader
            self.loader_registry["path"] = PathResourceLoader
            self.loader_registry["cli"] = CLIResourceLoader
            self.loader_registry["source"] = SourceResourceLoader
            self.loader_registry["callable"] = CallableResourceLoader
            self.loader_registry["image"] = ImageResourceLoader

        # Create MCP server
        self.server = Server(name)
        self._setup_handlers()
        self._setup_observers()
        logger.debug("Server initialized with name: %s", name)
        self._tasks: set[asyncio.Task[Any]] = set()

    def _create_task(self, coro: Coroutine[None, None, Any]) -> asyncio.Task[Any]:
        """Create and track an asyncio task."""
        task: asyncio.Task[Any] = asyncio.create_task(coro)
        self._tasks.add(task)
        task.add_done_callback(self._tasks.discard)
        return task

    def _setup_observers(self) -> None:
        """Set up registry observers."""
        # Create observers
        self.resource_observer = ResourceObserver(self)
        self.prompt_observer = PromptObserver(self)
        self.tool_observer = ToolObserver(self)

        # Register with registries
        self.resource_registry.add_observer(self.resource_observer.events)
        self.prompt_registry.add_observer(self.prompt_observer.events)
        self.tool_registry.add_observer(self.tool_observer.events)

    def _setup_handlers(self) -> None:
        """Register MCP protocol handlers."""

        @self.server.set_logging_level()
        async def handle_set_level(level: mcp.LoggingLevel) -> None:
            """Handle logging level changes."""
            # Map MCP levels to Python logging levels
            level_map = {
                "debug": logging.DEBUG,
                "info": logging.INFO,
                "notice": logging.INFO,
                "warning": logging.WARNING,
                "error": logging.ERROR,
                "critical": logging.CRITICAL,
                "alert": logging.CRITICAL,
                "emergency": logging.CRITICAL,
            }
            try:
                python_level = level_map[level]
                logger.setLevel(python_level)

                # Notify via log message
                await self.current_session.send_log_message(
                    level="info",
                    data=f"Log level set to {level}",
                    logger=self.name,
                )
            except Exception as exc:
                raise mcp.McpError(INTERNAL_ERROR, str(exc)) from exc

        @self.server.list_tools()
        async def handle_list_tools() -> list[mcp.types.Tool]:
            """Handle tools/list request."""
            return [conversions.to_mcp_tool(tool) for tool in self.tool_registry.values()]

        @self.server.call_tool()
        async def handle_call_tool(
            name: str, arguments: dict[str, Any] | None = None
        ) -> list[TextContent]:
            """Handle tools/call request."""
            try:
                result = await self.tool_registry.execute(name, **(arguments or {}))
                return [TextContent(type="text", text=str(result))]
            except Exception as exc:
                logger.exception("Tool execution failed: %s", name)
                error_msg = f"Tool execution failed: {exc}"
                return [TextContent(type="text", text=error_msg)]

        @self.server.list_prompts()
        async def handle_list_prompts() -> list[mcp.types.Prompt]:
            """Handle prompts/list request."""
            return [
                conversions.to_mcp_prompt(prompt)
                for prompt in self.prompt_registry.values()
            ]

        @self.server.get_prompt()
        async def handle_get_prompt(
            name: str, arguments: dict[str, str] | None = None
        ) -> GetPromptResult:
            """Handle prompts/get request."""
            result = await self.prompt_registry.render(name, arguments or {})
            messages = [conversions.to_mcp_message(msg) for msg in result.messages]
            return GetPromptResult(description=f"Prompt: {name}", messages=messages)

        @self.server.list_resources()
        async def handle_list_resources() -> list[mcp.types.Resource]:
            """Handle resources/list request."""
            resources = []

            # Use registry directly since we already registered resources in startup
            for name in self.resource_registry:
                try:
                    loaded = await self.resource_registry.load(name)
                    resources.append(conversions.to_mcp_resource(loaded))
                except Exception:
                    logger.exception("Failed to load resource %r", name)
                    continue

            logger.debug("Found %d resources", len(resources))
            return resources

        @self.server.read_resource()
        async def handle_read_resource(uri: mcp.types.AnyUrl) -> str | bytes:
            """Handle direct resource content requests."""
            try:
                # Convert MCP URI to our format and load
                internal_uri = conversions.from_mcp_uri(str(uri))
                logger.debug("Loading resource from internal URI: %s", internal_uri)

                # For plain resource names, load by name first
                if "://" not in internal_uri:
                    resource = await self.resource_registry.load(internal_uri)
                else:
                    resource = await self.resource_registry.load_by_uri(internal_uri)

                # Handle different content types
                if resource.metadata.mime_type.startswith("text/"):
                    return resource.content
                # For binary content (like images), return raw bytes
                return resource.content_items[0].content.encode()

            except Exception as exc:
                error_msg = f"Failed to read resource: {exc}"
                logger.exception(error_msg)
                # Use proper MCP error response
                raise mcp.McpError(mcp.types.INTERNAL_ERROR, error_msg) from exc

        @self.server.completion()
        async def handle_completion(
            ref: mcp.types.PromptReference | mcp.types.ResourceReference,
            argument: mcp.types.CompletionArgument,
        ) -> CompleteResult:
            """Handle completion requests."""
            try:
                completion: Completion
                match ref:
                    case mcp.types.PromptReference():
                        completion = await self._complete_prompt_argument(
                            self.prompt_registry[ref.name],
                            argument.name,
                            argument.value,
                        )
                    case mcp.types.ResourceReference():
                        completion = await self._complete_resource(
                            AnyUrl(ref.uri), argument
                        )
                    case _:
                        raise mcp.McpError(  # noqa: TRY301
                            INVALID_PARAMS, f"Invalid reference type: {type(ref)}"
                        )

                return CompleteResult(completion=completion)

            except Exception as exc:
                logger.exception("Completion failed")
                raise mcp.McpError(INTERNAL_ERROR, str(exc)) from exc

        @self.server.progress_notification()
        async def handle_progress(
            token: str | int,
            progress: float,
            total: float | None,
        ) -> None:
            """Handle progress notifications from client."""
            logger.debug(
                "Progress notification: %s %.1f/%.1f",
                token,
                progress,
                total or 0.0,
            )

    @classmethod
    def from_config_file(
        cls, config_path: str | os.PathLike[str], *, name: str = "llmling-server"
    ) -> LLMLingServer:
        """Create server from config file."""
        from llmling.config.loading import load_config

        return cls(load_config(config_path), name=name)

    async def start(self, *, raise_exceptions: bool = False) -> None:
        """Start the server."""
        try:
            # Initialize registries
            import logfire

            logfire.configure()
            await self.processor_registry.startup()
            await self.tool_registry.startup()
            await self.resource_registry.startup()

            # Register tools from config
            for name, tool_config in self.config.tools.items():
                self.tool_registry[name] = tool_config

            # Register resources from config
            for name, resource in self.config.resources.items():
                self.resource_registry.register(name, resource)
            # Start MCP server
            handler = configure_server_logging(self.server)
            options = self.server.create_initialization_options()
            async with (
                mcp.stdio_server() as (read_stream, write_stream),
                asyncio.TaskGroup() as tg,
            ):
                tg.create_task(run_logging_processor(handler))
                # Run server
                await self.server.run(
                    read_stream,
                    write_stream,
                    options,
                    raise_exceptions=raise_exceptions,
                )
        finally:
            await self.shutdown()

    async def shutdown(self) -> None:
        """Shutdown the server."""
        try:
            # Cancel all pending tasks
            if self._tasks:
                for task in self._tasks:
                    task.cancel()
                await asyncio.gather(*self._tasks, return_exceptions=True)

            # Remove observers
            if hasattr(self, "resource_observer"):
                await self.resource_observer.cleanup()
            if hasattr(self, "prompt_observer"):
                await self.prompt_observer.cleanup()
            if hasattr(self, "tool_observer"):
                await self.tool_observer.cleanup()
            self.resource_registry.remove_observer(self.resource_observer.events)
            self.prompt_registry.remove_observer(self.prompt_observer.events)
            self.tool_registry.remove_observer(self.tool_observer.events)

            # Shutdown registries
            await self.processor_registry.shutdown()
            await self.tool_registry.shutdown()
            await self.resource_registry.shutdown()

        finally:
            self._tasks.clear()

    async def __aenter__(self) -> Self:
        """Async context manager entry."""
        return self

    async def __aexit__(self, *exc: object) -> None:
        """Async context manager exit."""
        await self.shutdown()

    @property
    def current_session(self) -> mcp.ServerSession:
        """Get current session from request context."""
        try:
            return self.server.request_context.session
        except LookupError as exc:
            msg = "No active request context"
            raise RuntimeError(msg) from exc

    def notify_progress(
        self,
        token: str,
        progress: float,
        total: float | None = None,
        description: str | None = None,
    ) -> None:
        """Send progress notification to client."""
        try:
            # Get current session
            session = self.current_session

            # Create and track the progress notification task
            self._create_task(
                session.send_progress_notification(
                    progress_token=token,
                    progress=progress,
                    total=total,
                )
            )

            # Optionally send description as log message
            if description:
                self._create_task(
                    session.send_log_message(
                        level="info",
                        data=description,
                    )
                )

        except Exception:
            logger.exception("Failed to send progress notification")

    async def notify_resource_list_changed(self) -> None:
        """Notify clients about resource list changes."""
        try:
            await self.current_session.send_resource_list_changed()
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send resource list change notification")

    async def notify_resource_change(self, uri: str) -> None:
        """Notify clients about resource changes."""
        try:
            url = conversions.to_mcp_uri(uri)
            self._create_task(self.current_session.send_resource_updated(url))
            self._create_task(self.current_session.send_resource_list_changed())
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send resource change notification")

    async def notify_prompt_list_changed(self) -> None:
        """Notify clients about prompt list changes."""
        try:
            self._create_task(self.current_session.send_prompt_list_changed())
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send prompt list change notification")

    async def notify_tool_list_changed(self) -> None:
        """Notify clients about tool list changes."""
        try:
            self._create_task(self.current_session.send_tool_list_changed())
        except RuntimeError:
            logger.debug("No active session for notification")
        except Exception:
            logger.exception("Failed to send tool list change notification")

    async def _complete_prompt_argument(
        self,
        prompt: Prompt,
        arg_name: str,
        current_value: str,
    ) -> Completion:
        """Generate completions for prompt arguments."""
        # Find the argument definition
        arg_def = next(
            (arg for arg in prompt.arguments if arg.name == arg_name),
            None,
        )
        if not arg_def:
            return Completion(values=[], total=0, hasMore=False)

        values: list[str] = []
        match arg_def.type:
            case ArgumentType.ENUM:
                # Filter enum values based on current input
                values = [
                    v for v in (arg_def.enum_values or []) if v.startswith(current_value)
                ]

            case ArgumentType.RESOURCE:
                # Complete resource names
                values = [
                    name
                    for name in self.resource_registry.list_items()
                    if name.startswith(current_value)
                ]

            case ArgumentType.FILE:
                # Complete file paths
                import glob

                pattern = f"{current_value}*"
                values = list(glob.glob(pattern))  # noqa: PTH207

        return Completion(
            values=values[:100],  # Limit to 100 items
            total=len(values),
            hasMore=len(values) > 100,  # noqa: PLR2004
        )

    async def _complete_resource(
        self,
        uri: AnyUrl,
        argument: CompletionArgument,
    ) -> Completion:
        """Generate completions for resource fields."""
        try:
            # Convert AnyUrl to string for resource lookup
            str_uri = str(uri)
            resource = await self.resource_registry.load_by_uri(str_uri)
            values: list[str] = []

            # Different completion logic based on resource type
            match resource:
                case PathResource():
                    # Complete paths
                    import glob

                    pattern = f"{argument.value}*"
                    values = list(glob.glob(pattern))  # noqa: PTH207

                case SourceResource():
                    # Complete Python import paths
                    values = [
                        name
                        for name in self._get_importable_names()
                        if name.startswith(argument.value)
                    ]

            return Completion(
                values=values[:100],
                total=len(values),
                hasMore=len(values) > 100,  # noqa: PLR2004
            )

        except Exception:
            logger.exception("Resource completion failed")
            return Completion(values=[], total=0, hasMore=False)


async def serve(config_path: str | os.PathLike[str] | None = None) -> None:
    """Serve LLMling via MCP protocol.

    Args:
        config_path: Optional path to config file
    """
    logger = get_logger(__name__)

    # Create server instance
    server = LLMLingServer.from_config_file(config_path or config_resources.TEST_CONFIG)

    try:
        await server.start(raise_exceptions=True)
    except Exception:
        logger.exception("Server error")
        raise


if __name__ == "__main__":
    import asyncio
    import sys

    config_path = sys.argv[1] if len(sys.argv) > 1 else config_resources.TEST_CONFIG
    asyncio.run(serve(config_path))



from __future__ import annotations

import asyncio
import sys

from llmling.server import serve


if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


async def main() -> None:
    """Run the LLMling server."""
    config_path = (
        sys.argv[1] if len(sys.argv) > 1 else "src/llmling/config_resources/test.yml"
    )

    try:
        await serve(config_path)
    except KeyboardInterrupt:
        print("\nServer stopped by user", file=sys.stderr)
    except Exception as exc:  # noqa: BLE001
        print(f"Fatal server error: {exc}", file=sys.stderr)
        sys.exit(1)


def run() -> None:
    """Entry point for the server."""
    asyncio.run(main())


if __name__ == "__main__":
    run()



"""Function-based processor implementation."""

from __future__ import annotations

import asyncio
import inspect
from typing import TYPE_CHECKING, Any

import logfire

from llmling.core import exceptions
from llmling.core.log import get_logger
from llmling.processors.base import ChainableProcessor, ProcessorConfig, ProcessorResult
from llmling.utils import importing


if TYPE_CHECKING:
    from collections.abc import Callable

    from llmling.resources.models import ProcessingContext


logger = get_logger(__name__)


class FunctionProcessor(ChainableProcessor):
    """Processor that executes a Python function."""

    def __init__(self, config: ProcessorConfig) -> None:
        """Initialize processor."""
        super().__init__(config)
        self.func_config = config.get_function_config()
        self.func: Callable[..., Any] | None = None

    async def startup(self) -> None:
        """Load function during startup."""
        if not self.config.import_path:
            msg = "Import path not configured"
            raise exceptions.ProcessorError(msg)

        try:
            self.func = importing.import_callable(self.config.import_path)
        except ValueError as exc:
            msg = f"Failed to load function: {exc}"
            raise exceptions.ProcessorError(msg) from exc

    @logfire.instrument("Executing function processor")
    async def _process_impl(self, context: ProcessingContext) -> ProcessorResult:
        """Execute function with content."""
        if not self.func:
            msg = "Processor not initialized"
            raise exceptions.ProcessorError(msg)

        try:
            # Execute function
            result = self.func(context.current_content, **context.kwargs)

            # Handle async functions
            if inspect.iscoroutine(result):
                result = await result

            # Convert result to string
            content = str(result)

            return ProcessorResult(
                content=content,
                original_content=context.original_content,
                metadata={
                    "function": self.config.import_path,
                    "is_async": asyncio.iscoroutinefunction(self.func),
                },
            )
        except Exception as exc:
            msg = f"Function execution failed: {exc}"
            raise exceptions.ProcessorError(msg) from exc



"""Base classes for content processors."""

from __future__ import annotations

from abc import abstractmethod
from typing import TYPE_CHECKING, Any, Literal, TypedDict

from pydantic import BaseModel, ConfigDict, Field, model_validator

from llmling.core import exceptions
from llmling.core.log import get_logger


PROCESSOR_TYPES = Literal["function", "template", "image"]


if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from llmling.resources.models import ProcessingContext


logger = get_logger(__name__)


class FunctionProcessorConfig(TypedDict, total=False):
    """Configuration specific to function processors."""

    import_path: str
    async_execution: bool


class TemplateProcessorConfig(TypedDict, total=False):
    """Configuration specific to template processors."""

    template: str
    template_engine: Literal["jinja2"]


class ProcessorConfig(BaseModel):
    """Configuration for text processors."""

    type: PROCESSOR_TYPES
    name: str | None = None
    description: str | None = None

    # Function processor fields
    import_path: str = ""  # Required for function type
    async_execution: bool = False

    # Template processor fields
    template: str = ""  # Required for template type
    template_engine: Literal["jinja2"] = "jinja2"

    # Validation settings
    validate_output: bool = False
    validate_schema: dict[str, Any] | None = None

    # Additional settings
    timeout: float | None = None
    cache_results: bool = False
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(extra="forbid")

    # MCP-specific fields
    supported_mime_types: list[str] = Field(default_factory=lambda: ["text/plain"])
    max_input_size: int | None = None
    streaming: bool = False

    @model_validator(mode="after")
    def validate_config(self) -> ProcessorConfig:
        """Validate processor configuration based on type."""
        match self.type:
            case "function":
                if not self.import_path:
                    msg = "import_path is required for function processors"
                    raise ValueError(msg)
                self.name = self.name or self.import_path.split(".")[-1]
            case "template":
                if not self.template:
                    msg = "template is required for template processors"
                    raise ValueError(msg)
                self.name = self.name or "template_processor"
            case _:
                msg = f"Invalid processor type: {self.type}"
                raise ValueError(msg)
        return self

    def get_function_config(self) -> FunctionProcessorConfig:
        """Get function processor specific configuration."""
        if self.type != "function":
            msg = "Not a function processor configuration"
            raise ValueError(msg)
        return {
            "import_path": self.import_path,
            "async_execution": self.async_execution,
        }

    def get_template_config(self) -> TemplateProcessorConfig:
        """Get template processor specific configuration."""
        if self.type != "template":
            msg = "Not a template processor configuration"
            raise ValueError(msg)
        return {
            "template": self.template,
            "template_engine": self.template_engine,
        }


class ProcessorResult(BaseModel):
    """Result of processing content."""

    content: str
    original_content: str
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True)


class BaseProcessor:
    """Base class for all processors."""

    def __init__(self, config: ProcessorConfig) -> None:
        """Initialize processor with configuration."""
        self.config = config
        self._initialized = False

    async def startup(self) -> None:
        """Perform any necessary initialization."""
        self._initialized = True

    @abstractmethod
    async def process(self, context: ProcessingContext) -> ProcessorResult:
        """Process content with given context.

        Args:
            context: Processing context

        Returns:
            Processing result

        Raises:
            ProcessorError: If processing fails
        """

    async def shutdown(self) -> None:
        """Perform any necessary cleanup."""

    async def validate_result(self, result: ProcessorResult) -> None:
        """Validate processing result."""
        if not self.config.validate_output:
            return

        if not result.content:
            msg = "Processor returned empty content"
            raise exceptions.ProcessorError(msg)

        if self.config.validate_schema:  # noqa: SIM102
            # Schema validation would go here
            if not isinstance(result.content, str):
                msg = f"Expected string output, got {type(result.content)}"
                raise exceptions.ProcessorError(msg)


class AsyncProcessor(BaseProcessor):
    """Base class for asynchronous processors."""

    async def process_stream(
        self,
        context: ProcessingContext,
    ) -> AsyncIterator[ProcessorResult]:
        """Process content in streaming mode."""
        result = await self.process(context)
        yield result


class ChainableProcessor(AsyncProcessor):
    """Processor that can be chained with others."""

    async def pre_process(self, context: ProcessingContext) -> ProcessingContext:
        """Prepare context for processing."""
        return context

    async def post_process(
        self,
        context: ProcessingContext,
        result: ProcessorResult,
    ) -> ProcessorResult:
        """Modify result after processing."""
        return result

    async def process(self, context: ProcessingContext) -> ProcessorResult:
        """Process content with pre and post processing."""
        try:
            prepared_context = await self.pre_process(context)
            result = await self._process_impl(prepared_context)
            final_result = await self.post_process(prepared_context, result)
            await self.validate_result(final_result)
        except Exception as exc:
            msg = f"Processing failed: {exc}"
            raise exceptions.ProcessorError(msg) from exc
        else:
            return final_result

    @abstractmethod
    async def _process_impl(self, context: ProcessingContext) -> ProcessorResult:
        """Implement actual processing logic."""



from llmling.prompts.manager import PromptManager
from llmling.prompts.models import MessageContext, SystemPrompt
from llmling.prompts.models import (
    Prompt,
    ExtendedPromptArgument,
    PromptMessage,
    PromptResult,
)
from llmling.prompts.registry import PromptRegistry
from llmling.prompts.rendering import render_prompt

__all__ = [
    "ExtendedPromptArgument",
    "MessageContext",
    "Prompt",
    "PromptManager",
    "PromptMessage",
    "PromptRegistry",
    "PromptResult",
    "SystemPrompt",
    "render_prompt",
]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar

from llmling.config.models import CallableResource
from llmling.core import exceptions
from llmling.core.log import get_logger
from llmling.resources.base import ResourceLoader, create_loaded_resource
from llmling.utils import calling


if TYPE_CHECKING:
    from llmling.processors.registry import ProcessorRegistry
    from llmling.resources.models import LoadedResource


logger = get_logger(__name__)


class CallableResourceLoader(ResourceLoader[CallableResource]):
    """Loads context from Python callable execution."""

    context_class = CallableResource
    uri_scheme = "callable"
    supported_mime_types: ClassVar[list[str]] = ["text/plain"]

    async def _load_impl(
        self,
        resource: CallableResource,
        name: str,
        processor_registry: ProcessorRegistry | None,
    ) -> LoadedResource:
        """Execute callable and load result."""
        try:
            content = await calling.execute_callable(
                resource.import_path,
                **resource.keyword_args,
            )

            if processor_registry and (procs := resource.processors):
                processed = await processor_registry.process(content, procs)
                content = processed.content

            return create_loaded_resource(
                content=content,
                source_type="callable",
                uri=self.create_uri(name=name),
                name=resource.description or resource.import_path,
                description=resource.description,
                additional_metadata={
                    "import_path": resource.import_path,
                    "args": resource.keyword_args,
                },
            )
        except Exception as exc:
            msg = f"Failed to execute callable {resource.import_path}"
            raise exceptions.LoaderError(msg) from exc



"""Tool system for LLMling."""

from __future__ import annotations

from llmling.tools.base import LLMCallableTool
from llmling.tools.registry import ToolRegistry
from llmling.tools.exceptions import ToolError, ToolExecutionError

__all__ = [
    "LLMCallableTool",
    "ToolError",
    "ToolExecutionError",
    "ToolRegistry",
]



"""Event types for registry observers."""

from __future__ import annotations

from typing import TYPE_CHECKING, Generic, TypeVar


if TYPE_CHECKING:
    from collections.abc import Callable

    from llmling.core.typedefs import MetadataDict


TKey = TypeVar("TKey")
TItem = TypeVar("TItem")


class RegistryEvents(Generic[TKey, TItem]):
    """Event callbacks for registry changes."""

    def __init__(self) -> None:
        """Initialize empty callbacks."""
        self.on_item_added: Callable[[TKey, TItem], None] | None = None
        self.on_item_removed: Callable[[TKey, TItem], None] | None = None
        self.on_item_modified: Callable[[TKey, TItem], None] | None = None
        self.on_list_changed: Callable[[], None] | None = None
        self.on_reset: Callable[[], None] | None = None

    def __repr__(self) -> str:
        """Show which callbacks are set."""
        callbacks = []
        for name, cb in vars(self).items():
            if cb is not None:
                callbacks.append(name)
        return f"{self.__class__.__name__}(active={callbacks})"


class ResourceEvents(RegistryEvents[str, "Resource"]):  # type: ignore
    """Resource-specific registry events."""

    def __init__(self) -> None:
        """Initialize with resource-specific callbacks."""
        super().__init__()
        self.on_content_changed: Callable[[str, str | bytes], None] | None = None
        self.on_metadata_changed: Callable[[str, MetadataDict], None] | None = None



from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING

import anyio
from mcp import types
import pytest

from llmling.config.manager import ConfigManager
from llmling.config.models import Config, GlobalSettings, TextResource, ToolConfig
from llmling.processors.registry import ProcessorRegistry
from llmling.prompts.models import (
    ArgumentType,
    ExtendedPromptArgument,
    Prompt,
    PromptMessage,
)
from llmling.server import LLMLingServer


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator


@pytest.fixture(autouse=True)
def disable_logfire():
    """Disable Logfire for all tests."""
    import os

    os.environ["LOGFIRE_IGNORE_NO_CONFIG"] = "1"


@pytest.fixture
def config_manager(test_config):
    """Get config manager with test configuration."""
    return ConfigManager(test_config)


@pytest.fixture
def processor_registry():
    """Get clean processor registry."""
    return ProcessorRegistry()


@pytest.fixture
def test_config() -> Config:
    """Create test configuration with sample data."""
    prompt_msg = PromptMessage(role="user", content="Test message")
    return Config(
        version="1.0.0",
        global_settings=GlobalSettings(),
        resources={
            "test-resource": TextResource(
                content="test content", description="Test resource"
            ),
        },
        tools={
            "example": ToolConfig(
                import_path="llmling.testing.tools.example_tool",
                name="example",
                description="Example tool for testing",
            ),
        },
        prompts={
            "test-prompt": Prompt(
                name="test-prompt",
                description="Test prompt",
                messages=[prompt_msg],
            ),
        },
    )


@pytest.fixture
async def server(test_config: Config) -> AsyncGenerator[LLMLingServer, None]:
    """Create and initialize a test server."""
    server = LLMLingServer(test_config)

    # Register test prompts
    server.prompt_registry.register(
        "test-prompt",
        Prompt(
            name="test-prompt",
            description="Test prompt with completable arguments",
            arguments=[
                ExtendedPromptArgument(
                    name="mode",
                    description="Operation mode",
                    type=ArgumentType.ENUM,
                    enum_values=["quick", "detailed", "advanced"],
                ),
                ExtendedPromptArgument(
                    name="tool",
                    description="Tool to use",
                    type=ArgumentType.TOOL,
                    tool_names=["*"],
                ),
                ExtendedPromptArgument(
                    name="input_file",
                    description="Input file path",
                    type=ArgumentType.FILE,
                    file_patterns=["*.txt", "*.md"],
                ),
            ],
            messages=[],
        ),
    )

    # Create memory streams for the server
    send_stream, receive_stream = anyio.create_memory_object_stream[
        types.JSONRPCMessage
    ]()

    # Start the server's MCP server in the background
    server_task = asyncio.create_task(
        server.server.run(
            receive_stream,
            send_stream,
            server.server.create_initialization_options(),
        )
    )

    await server.start()
    try:
        yield server
    finally:
        server_task.cancel()
        await server.shutdown()



from __future__ import annotations

import asyncio
import contextlib
import json
import os
import subprocess
import sys
from typing import Any

from llmling.core.log import get_logger


logger = get_logger(__name__)


class TestClient:
    def __init__(self, server_command: list[str]) -> None:
        self.server_command = server_command
        self.process: subprocess.Popen[bytes] | None = None

    async def start(self) -> None:
        """Start the server process."""
        env = {**os.environ, "PYTHONUNBUFFERED": "1"}
        self.process = subprocess.Popen(
            self.server_command,
            stdin=subprocess.PIPE,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
            bufsize=0,
            env=env,
        )
        logger.debug("Started server process")
        await asyncio.sleep(0.5)

    async def _read_response(self) -> dict[str, Any]:
        """Read a response from the server."""
        if not self.process or not self.process.stdout:
            msg = "Server process not available"
            raise RuntimeError(msg)

        # Also read stderr in background to see server logs
        async def read_stderr():
            while True:
                if not self.process or not self.process.stderr:
                    break
                err = await asyncio.get_event_loop().run_in_executor(
                    None, self.process.stderr.readline
                )
                if not err:
                    break
                logger.debug("Server stderr: %s", err.decode().strip())

        # Start stderr reader and store the task
        self._stderr_task = asyncio.create_task(read_stderr())

        # Read stdout
        line = await asyncio.get_event_loop().run_in_executor(
            None, self.process.stdout.readline
        )

        if not line:
            msg = "Server closed connection"
            raise RuntimeError(msg)

        try:
            response = json.loads(line.decode())
            logger.debug("Received JSON: %s", response)
        except json.JSONDecodeError:
            # Log non-JSON lines as debug messages
            logger.debug("Server stdout: %s", line.decode().strip())
            # Keep reading until we get JSON
            return await self._read_response()
        else:
            return response

    async def send_request(
        self, method: str, params: dict[str, Any] | None = None
    ) -> Any:
        """Send JSON-RPC request to server."""
        if not self.process or not self.process.stdin:
            msg = "Server not started"
            raise RuntimeError(msg)

        request = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params or {},
            "id": 1,
        }

        # Send request
        request_str = json.dumps(request) + "\n"
        logger.debug("Sending: %s", request_str.strip())
        self.process.stdin.write(request_str.encode())
        self.process.stdin.flush()

        # Wait for response with timeout
        try:
            async with asyncio.timeout(5.0):
                while True:
                    response = await self._read_response()
                    if "id" in response and response["id"] == request["id"]:
                        if "error" in response:
                            msg = f"Server error: {response['error']}"
                            raise RuntimeError(msg)
                        return response.get("result")
        except TimeoutError:
            msg = "Timeout waiting for server response"
            logger.exception(msg)
            raise

    async def send_notification(
        self, method: str, params: dict[str, Any] | None = None
    ) -> None:
        """Send notification (no response expected)."""
        if not self.process or not self.process.stdin:
            msg = "Server not started"
            raise RuntimeError(msg)

        notification = {
            "jsonrpc": "2.0",
            "method": method,
            "params": params or {},
        }

        notification_str = json.dumps(notification) + "\n"
        logger.debug("Sending notification: %s", notification_str.strip())
        self.process.stdin.write(notification_str.encode())
        self.process.stdin.flush()

    async def list_tools(self) -> list[dict[str, Any]]:
        """List available tools."""
        logger.debug("Starting tools listing sequence")

        # First initialize
        logger.debug("Sending initialize request")
        response = await self.send_request(
            "initialize",
            {
                "protocolVersion": "0.1",
                "capabilities": {},
                "clientInfo": {"name": "test-client", "version": "1.0"},
                "processId": None,
                "rootUri": None,
                "workspaceFolders": None,
            },
        )
        logger.debug("Initialize response: %s", response)

        # Send initialized notification
        logger.debug("Sending initialized notification")
        await self.send_notification("notifications/initialized", {})

        # Now list tools
        logger.debug("Sending tools/list request")
        return await self.send_request("tools/list")

    async def close(self) -> None:
        """Stop the server."""
        if hasattr(self, "_stderr_task"):
            self._stderr_task.cancel()
            with contextlib.suppress(asyncio.CancelledError):
                await self._stderr_task

        if self.process:
            try:
                await self.send_request("shutdown", {})
                await self.send_notification("exit", {})
            except Exception as e:  # noqa: BLE001
                logger.warning("Error during shutdown: %s", e)
            finally:
                self.process.terminate()
                try:
                    self.process.wait(timeout=1.0)
                except subprocess.TimeoutExpired:
                    self.process.kill()


async def main() -> None:
    import logging

    logging.basicConfig(
        level=logging.DEBUG, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
    )

    client = TestClient([sys.executable, "-m", "llmling.server"])
    try:
        await client.start()
        tools = await client.list_tools()
        print("\nAvailable tools:", tools)
    except Exception:
        logger.exception("Error")
    finally:
        await client.close()


if __name__ == "__main__":
    asyncio.run(main())



"""Registry for prompt templates."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from llmling.core import exceptions
from llmling.core.baseregistry import BaseRegistry
from llmling.prompts.models import Prompt, PromptResult
from llmling.prompts.rendering import render_prompt


if TYPE_CHECKING:
    from llmling.processors.registry import ProcessorRegistry
    from llmling.resources.loaders.registry import ResourceLoaderRegistry


class PromptRegistry(BaseRegistry[str, Prompt]):
    """Registry for prompt templates."""

    def __init__(
        self,
        resource_registry: ResourceLoaderRegistry | None = None,
        processor_registry: ProcessorRegistry | None = None,
    ) -> None:
        """Initialize registry.

        Args:
            resource_registry: Optional registry for resolving resources
            processor_registry: Optional registry for content processors
        """
        super().__init__()
        self.resource_registry = resource_registry
        self.processor_registry = processor_registry

    @property
    def _error_class(self) -> type[exceptions.LLMLingError]:
        return exceptions.LLMLingError

    def _validate_item(self, item: Any) -> Prompt:
        """Validate and convert items to Prompt instances."""
        match item:
            case Prompt():
                return item
            case dict():
                return Prompt.model_validate(item)
            case _:
                msg = f"Invalid prompt type: {type(item)}"
                raise exceptions.LLMLingError(msg)

    async def render(
        self,
        name: str,
        arguments: dict[str, Any] | None = None,
    ) -> PromptResult:
        """Render a prompt template with arguments."""
        prompt = self[name]
        return await render_prompt(
            prompt,
            arguments or {},
            resource_registry=self.resource_registry,
            processor_registry=self.processor_registry,
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from pydantic import ValidationError
import pytest

from llmling.config.models import Config
from llmling.core.exceptions import ProcessorError
from llmling.prompts.models import (
    ArgumentType,
    ExtendedPromptArgument,
    MessageContent,
    Prompt,
    PromptMessage,
    PromptResult,
)
from llmling.prompts.rendering import render_prompt
from llmling.resources.base import LoaderContext, ResourceLoader
from llmling.resources.models import LoadedResource, ResourceMetadata


if TYPE_CHECKING:
    from llmling.processors.registry import ProcessorRegistry


def test_config_with_prompts():
    """Test config with prompts section."""
    config_data = {
        "version": "1.0",
        # Need to add required fields from Config model
        "resources": {},
        "prompts": {
            "analyze": {
                "name": "analyze",
                "description": "Analyze code",
                "messages": [{"role": "user", "content": "Analyze this code: {code}"}],
                "arguments": [
                    {
                        "name": "code",
                        "type": "text",
                        "description": "Code to analyze",
                        "required": True,
                    }
                ],
            }
        },
    }
    config = Config.model_validate(config_data)
    assert "analyze" in config.prompts
    assert config.prompts["analyze"].name == "analyze"


def test_config_with_resource_prompts():
    """Test config with prompts using resources."""
    config_data = {
        "version": "1.0",
        "resources": {},  # Required field
        "prompts": {
            "review": {
                "name": "review",
                "description": "Review code and tests",
                "messages": [
                    {
                        "role": "user",
                        "content": [
                            MessageContent(
                                type="text", content="Review this implementation:"
                            ),
                            MessageContent(
                                type="resource",
                                content="source://code.py",
                                alt_text="Source code",
                            ),
                        ],
                    }
                ],
            }
        },
    }
    config = Config.model_validate(config_data)
    assert "review" in config.prompts
    msg = config.prompts["review"].messages[0]
    assert isinstance(msg.content, list)
    assert len(msg.content) == 2  # noqa: PLR2004
    assert msg.content[0].type == "text"
    assert msg.content[1].type == "resource"
    assert msg.content[1].content == "source://code.py"


def test_invalid_prompt_config():
    """Test invalid prompt configurations."""
    with pytest.raises(ValidationError):
        Config.model_validate({
            "version": "1.0",
            "prompts": {
                "invalid": {
                    "name": "invalid",
                    "messages": [
                        {
                            "role": "invalid_role",  # Invalid role
                            "content": "test",
                        }
                    ],
                }
            },
        })


@pytest.fixture
def sample_prompt() -> Prompt:
    """Create a sample prompt for testing."""
    arg = ExtendedPromptArgument(
        name="name",
        type=ArgumentType.TEXT,
        description="Name to greet",
        required=True,
    )
    return Prompt(
        name="test",
        description="Test prompt",
        messages=[PromptMessage(role="user", content="Hello {name}")],
        arguments=[arg],
    )


@pytest.fixture
def resource_prompt() -> Prompt:
    """Create a prompt with resource references."""
    return Prompt(
        name="analyze",
        description="Analyze code",
        messages=[
            PromptMessage(
                role="user",
                content=[
                    MessageContent.text("Analyze this code:"),
                    MessageContent.resource("source://test.py"),
                ],
            )
        ],
    )


@pytest.fixture
def loaded_resource() -> LoadedResource:
    """Create a sample loaded resource."""
    return LoadedResource(
        content="def test(): pass",
        source_type="source",
        metadata=ResourceMetadata(
            uri="source://test.py", mime_type="text/x-python", name="test.py"
        ),
    )


async def test_prompt_rendering(sample_prompt: Prompt):
    """Test basic prompt rendering."""
    result = await render_prompt(sample_prompt, {"name": "World"})
    assert isinstance(result, PromptResult)
    assert len(result.messages) == 1
    assert isinstance(result.messages[0].content, list)
    assert result.messages[0].content[0].content == "Hello World"


async def test_prompt_with_resources(
    resource_prompt: Prompt,
    loaded_resource: LoadedResource,
):
    """Test prompt rendering with resource resolution."""

    class MockProcessorRegistry:
        async def process(self, content: str, *args: Any, **kwargs: Any) -> str:
            return "processed content"

    class MockLoader(ResourceLoader[None]):
        uri_scheme = "source"

        async def _load_impl(
            self,
            resource: Any,
            name: str,
            processor_registry: ProcessorRegistry | None,
        ) -> LoadedResource:
            return loaded_resource

        async def load(
            self,
            context: LoaderContext[None] | None = None,
            processor_registry: ProcessorRegistry | None = None,
        ) -> LoadedResource:
            return loaded_resource

    class MockResourceRegistry:
        def find_loader_for_uri(self, uri: str) -> ResourceLoader[Any]:
            return MockLoader()

    result = await render_prompt(
        resource_prompt,
        {},
        resource_registry=MockResourceRegistry(),
        processor_registry=MockProcessorRegistry(),
    )
    assert len(result.messages) == 1
    msg = result.messages[0]
    assert isinstance(msg.content, list)
    assert len(msg.content) == 2  # noqa: PLR2004
    assert msg.content[0].type == "text"
    assert msg.content[0].content == "Analyze this code:"
    assert msg.content[1].type == "resource"
    assert msg.resolved_content is not None
    assert len(msg.resolved_content) == 2  # noqa: PLR2004
    assert msg.resolved_content[0].original == msg.content[0]
    assert msg.resolved_content[1].resolved == loaded_resource


async def test_prompt_validation():
    """Test prompt argument validation."""
    arg = ExtendedPromptArgument(
        name="required_arg",
        type=ArgumentType.TEXT,
        required=True,
    )
    prompt = Prompt(
        name="test",
        description="",
        messages=[PromptMessage(role="user", content="Test {required_arg}")],
        arguments=[arg],
    )

    with pytest.raises(ProcessorError, match="Missing required argument"):
        await render_prompt(prompt, {})



"""Observer implementations for converting registry events to MCP notifications."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from llmling.config.models import Resource
from llmling.core.events import RegistryEvents
from llmling.core.log import get_logger
from llmling.prompts.models import Prompt
from llmling.tools.base import LLMCallableTool


if TYPE_CHECKING:
    from collections.abc import Coroutine

    from llmling.server.server import LLMLingServer


logger = get_logger(__name__)


class ServerObserver:
    """Base observer with server reference and task tracking."""

    def __init__(self, server: LLMLingServer) -> None:
        """Initialize with server reference."""
        self.server = server

    def _create_notification_task(self, coro: Coroutine[None, None, Any]) -> None:
        """Create a tracked notification task."""
        self.server._create_task(coro)

    async def cleanup(self) -> None:
        """Wait for pending notifications to complete."""


class ResourceObserver(ServerObserver):
    """Converts resource registry events to MCP notifications."""

    def __init__(self, server: LLMLingServer) -> None:
        """Initialize and create events object."""
        super().__init__(server)
        self.events = RegistryEvents[str, Resource]()
        # Set up callbacks
        self.events.on_item_added = self._handle_resource_changed
        self.events.on_item_modified = self._handle_resource_changed
        self.events.on_item_removed = self._handle_list_changed
        self.events.on_reset = self._handle_list_changed

    def _handle_resource_changed(self, key: str, resource: Resource) -> None:
        """Handle individual resource changes."""
        try:
            loader = self.server.loader_registry.get_loader(resource)
            uri = loader.create_uri(name=key)
            self._create_notification_task(self.server.notify_resource_change(uri))
        except Exception:
            logger.exception("Failed to notify resource change")

    def _handle_list_changed(self, *args: object) -> None:
        """Handle changes that affect the resource list."""
        try:
            self._create_notification_task(self.server.notify_resource_list_changed())
        except Exception:
            logger.exception("Failed to notify list change")


class PromptObserver(ServerObserver):
    """Converts prompt registry events to MCP notifications."""

    def __init__(self, server: LLMLingServer) -> None:
        """Initialize and create events object."""
        super().__init__(server)
        self.events = RegistryEvents[str, Prompt]()
        # Any prompt change triggers list update
        self.events.on_item_added = self._handle_list_changed
        self.events.on_item_modified = self._handle_list_changed
        self.events.on_item_removed = self._handle_list_changed
        self.events.on_reset = self._handle_list_changed

    def _handle_list_changed(self, *args: object) -> None:
        """Handle any prompt changes."""
        try:
            self._create_notification_task(self.server.notify_prompt_list_changed())
        except Exception:
            logger.exception("Failed to notify prompt list change")


class ToolObserver(ServerObserver):
    """Converts tool registry events to MCP notifications."""

    def __init__(self, server: LLMLingServer) -> None:
        """Initialize and create events object."""
        super().__init__(server)
        self.events = RegistryEvents[str, LLMCallableTool]()
        # Any tool change triggers list update
        self.events.on_item_added = self._handle_list_changed
        self.events.on_item_modified = self._handle_list_changed
        self.events.on_item_removed = self._handle_list_changed
        self.events.on_reset = self._handle_list_changed

    def _handle_list_changed(self, *args: object) -> None:
        """Handle any tool changes."""
        try:
            self._create_notification_task(self.server.notify_tool_list_changed())
        except Exception:
            logger.exception("Failed to notify tool list change")



"""Registry for content processors."""

from __future__ import annotations

import asyncio
from typing import TYPE_CHECKING, Any

import logfire

from llmling.core import exceptions
from llmling.core.baseregistry import BaseRegistry
from llmling.core.log import get_logger
from llmling.processors.base import BaseProcessor, ProcessorConfig, ProcessorResult
from llmling.processors.implementations.function import FunctionProcessor
from llmling.processors.implementations.template import TemplateProcessor
from llmling.resources.models import ProcessingContext


if TYPE_CHECKING:
    from collections.abc import AsyncIterator

    from llmling.core.typedefs import ProcessingStep


logger = get_logger(__name__)


class ProcessorRegistry(BaseRegistry[str, BaseProcessor]):
    """Registry and execution manager for processors."""

    @property
    def _error_class(self) -> type[exceptions.ProcessorError]:
        return exceptions.ProcessorError

    def _validate_item(self, item: Any) -> BaseProcessor:
        """Validate and transform items."""
        match item:
            case BaseProcessor():
                return item
            case ProcessorConfig():
                return self._create_processor(item)
            case _ if callable(item):
                # Use existing FunctionProcessor
                config = ProcessorConfig(
                    type="function",
                    import_path=f"{item.__module__}.{item.__qualname__}",
                    async_execution=asyncio.iscoroutinefunction(item),
                )
                return FunctionProcessor(config)
            case _:
                msg = f"Invalid processor type: {type(item)}"
                raise exceptions.ProcessorError(msg)

    def _create_processor(self, config: ProcessorConfig) -> BaseProcessor:
        """Create processor instance from configuration."""
        try:
            match config.type:
                case "function":
                    return FunctionProcessor(config)
                case "template":
                    return TemplateProcessor(config)
                case _:
                    msg = f"Unknown processor type: {config.type}"
                    raise exceptions.ProcessorError(msg)  # noqa: TRY301
        except Exception as exc:
            msg = f"Failed to create processor for config {config}"
            raise exceptions.ProcessorError(msg) from exc

    async def get_processor(self, name: str) -> BaseProcessor:
        """Get a processor by name (backward compatibility)."""
        processor = self.get(name)
        if not getattr(processor, "_initialized", False):
            await processor.startup()
            processor._initialized = True  # type: ignore
        return processor

    @logfire.instrument("Processing content through steps")
    async def process(
        self,
        content: str,
        steps: list[ProcessingStep],
        metadata: dict[str, Any] | None = None,
    ) -> ProcessorResult:
        """Process content through steps."""
        if not self._initialized:
            await self.startup()

        current_context = ProcessingContext(
            original_content=content,
            current_content=content,
            metadata=metadata or {},
            kwargs={},
        )

        # Group parallel steps together
        parallel_groups: list[list[ProcessingStep]] = [[]]
        for step in steps:
            if step.parallel and parallel_groups[-1]:
                parallel_groups[-1].append(step)
            else:
                if parallel_groups[-1]:
                    parallel_groups.append([])
                parallel_groups[-1].append(step)

        # Process each group
        result = None
        for group in parallel_groups:
            if len(group) > 1:  # Parallel processing
                try:
                    result = await self.process_parallel_steps(group, current_context)
                except Exception as exc:
                    # All parallel steps failed
                    msg = f"All parallel steps failed: {exc}"
                    raise exceptions.ProcessorError(msg) from exc
            else:  # Sequential processing
                step = group[0]
                step_context = ProcessingContext(
                    original_content=current_context.original_content,
                    current_content=current_context.current_content,
                    metadata=current_context.metadata,
                    kwargs=step.kwargs or {},
                )

                processor = await self.get_processor(step.name)
                try:
                    result = await processor.process(step_context)
                except Exception as exc:
                    if step.required:
                        msg = f"Required step {step.name} failed: {exc}"
                        raise exceptions.ProcessorError(msg) from exc

                    # Optional step failed, continue with current context
                    logger.warning(
                        "Optional step %s failed: %s",
                        step.name,
                        exc,
                    )
                    result = ProcessorResult(
                        content=current_context.current_content,
                        original_content=current_context.original_content,
                        metadata=current_context.metadata,
                    )

            # Update context for next group
            if result:
                current_context = ProcessingContext(
                    original_content=content,
                    current_content=result.content,
                    metadata={**current_context.metadata, **result.metadata},
                    kwargs={},
                )

        return (
            result
            if result
            else ProcessorResult(
                content=content,
                original_content=content,
                metadata=current_context.metadata,
            )
        )

    async def process_stream(
        self,
        content: str,
        steps: list[ProcessingStep],
        metadata: dict[str, Any] | None = None,
    ) -> AsyncIterator[ProcessorResult]:
        """Process content through steps, yielding intermediate results."""
        if not self._initialized:
            await self.startup()

        current_context = ProcessingContext(
            original_content=content,
            current_content=content,
            metadata=metadata or {},
            kwargs={},
        )

        for step in steps:
            processor = await self.get_processor(step.name)

            # Create new context with step kwargs
            step_context = ProcessingContext(
                original_content=current_context.original_content,
                current_content=current_context.current_content,
                metadata=current_context.metadata,
                kwargs=step.kwargs or {},
            )

            try:
                result = await processor.process(step_context)

                # Update context for next step
                current_context = ProcessingContext(
                    original_content=content,
                    current_content=result.content,
                    metadata={**current_context.metadata, **result.metadata},
                    kwargs={},
                )

                yield result
            except Exception as exc:
                if not step.required:
                    continue
                msg = f"Step {step.name} failed"
                raise exceptions.ProcessorError(msg) from exc

    async def process_parallel_steps(
        self,
        steps: list[ProcessingStep],
        context: ProcessingContext,
    ) -> ProcessorResult:
        """Process steps in parallel."""
        results: list[ProcessorResult] = []

        for step in steps:
            step_context = ProcessingContext(
                original_content=context.original_content,
                current_content=context.current_content,
                metadata=context.metadata,
                kwargs=step.kwargs or {},
            )

            processor = await self.get_processor(step.name)
            try:
                result = await processor.process(step_context)
                results.append(result)
            except Exception as exc:
                if step.required:
                    raise
                logger.warning(
                    "Optional parallel step %s failed: %s",
                    step.name,
                    exc,
                )

        if not results:
            # If all steps failed and were optional, return original context
            return ProcessorResult(
                content=context.current_content,
                original_content=context.original_content,
                metadata=context.metadata,
            )

        # Combine successful results
        combined_content = "\n".join(r.content for r in results)
        combined_metadata: dict[str, Any] = {}
        for result in results:
            combined_metadata.update(result.metadata)

        return ProcessorResult(
            content=combined_content,
            original_content=context.original_content,
            metadata=combined_metadata,
        )



from __future__ import annotations

from typing import Any

import pytest

from llmling.core import exceptions
from llmling.core.baseregistry import BaseRegistry
from llmling.core.events import RegistryEvents


# Simple test registry
class RegItem:
    """Test item for registry."""

    def __init__(self, value: str):
        self.value = value


class Registry(BaseRegistry[str, RegItem]):
    """Test registry implementation."""

    @property
    def _error_class(self) -> type[exceptions.LLMLingError]:
        return exceptions.LLMLingError

    def _validate_item(self, item: Any) -> RegItem:
        if isinstance(item, RegItem):
            return item
        if isinstance(item, str):
            return RegItem(item)
        msg = f"Invalid item type: {type(item)}"
        raise self._error_class(msg)


# Tests
def test_registry_observer_registration():
    """Test that observers can be registered and unregistered."""
    registry = Registry()
    events = RegistryEvents[str, RegItem]()

    # Register
    registry.add_observer(events)
    assert events in registry._observers

    # Unregister
    registry.remove_observer(events)
    assert events not in registry._observers


def test_item_added_event():
    """Test that item addition triggers event."""
    registry = Registry()
    events = RegistryEvents[str, RegItem]()

    # Track calls
    called_with = None

    def on_added(key: str, item: RegItem) -> None:
        nonlocal called_with
        called_with = (key, item)

    # Set up observer
    events.on_item_added = on_added
    registry.add_observer(events)

    # Add item
    test_item = RegItem("test")
    registry.register("test", test_item)

    # Check event was triggered
    assert called_with is not None
    key, item = called_with
    assert key == "test"
    assert item is test_item


def test_item_modified_event():
    """Test that item modification triggers event."""
    registry = Registry()
    events = RegistryEvents[str, RegItem]()

    # Track calls
    called_with = None

    def on_modified(key: str, item: RegItem) -> None:
        nonlocal called_with
        called_with = (key, item)

    # Set up observer
    events.on_item_modified = on_modified
    registry.add_observer(events)

    # Add and modify item
    registry.register("test", RegItem("original"))
    new_item = RegItem("modified")
    registry.register("test", new_item, replace=True)

    # Check event was triggered
    assert called_with is not None
    key, item = called_with
    assert key == "test"
    assert item is new_item


def test_item_removed_event():
    """Test that item removal triggers event."""
    registry = Registry()
    events = RegistryEvents[str, RegItem]()

    # Track calls
    called_with = None

    def on_removed(key: str, item: RegItem) -> None:
        nonlocal called_with
        called_with = (key, item)

    # Set up observer
    events.on_item_removed = on_removed
    registry.add_observer(events)

    # Add and remove item
    test_item = RegItem("test")
    registry.register("test", test_item)
    del registry["test"]

    # Check event was triggered
    assert called_with is not None
    key, item = called_with
    assert key == "test"
    assert item is test_item


def test_reset_event():
    """Test that reset triggers event."""
    registry = Registry()
    events = RegistryEvents[str, RegItem]()

    # Track calls
    was_called = False

    def on_reset() -> None:
        nonlocal was_called
        was_called = True

    # Set up observer
    events.on_reset = on_reset
    registry.add_observer(events)

    # Add item and reset
    registry.register("test", RegItem("test"))
    registry.reset()

    # Check event was triggered
    assert was_called


def test_multiple_observers():
    """Test that multiple observers receive events."""
    registry = Registry()
    events1 = RegistryEvents[str, RegItem]()
    events2 = RegistryEvents[str, RegItem]()

    # Track calls
    calls = []

    def on_added1(key: str, item: RegItem) -> None:
        calls.append(1)

    def on_added2(key: str, item: RegItem) -> None:
        calls.append(2)

    # Set up observers
    events1.on_item_added = on_added1
    events2.on_item_added = on_added2
    registry.add_observer(events1)
    registry.add_observer(events2)

    # Add item
    registry.register("test", RegItem("test"))

    # Check both observers were called
    assert len(calls) == 2  # noqa: PLR2004
    assert 1 in calls
    assert 2 in calls  # noqa: PLR2004


if __name__ == "__main__":
    pytest.main(["-vv", __file__])



"""Configuration management for LLMling."""

from __future__ import annotations

from llmling.config.manager import ConfigManager
from llmling.config.models import (
    Config,
    Resource,
    GlobalSettings,
    CallableResource,
    SourceResource,
)
from llmling.config.validation import ConfigValidator
from llmling.config.loading import load_config


__all__ = [
    "CallableResource",
    "Config",
    "ConfigManager",
    "ConfigValidator",
    "GlobalSettings",
    "Resource",
    "SourceResource",
    "load_config",
]



"""File system watching for resources."""

from llmling.resources.watching.models import WatchConfig
from llmling.resources.watching.watcher import ResourceWatcher
from llmling.resources.watching.utils import load_patterns

__all__ = [
    "ResourceWatcher",
    "WatchConfig",
    "load_patterns",
]



"""Tests for context loaders."""

from __future__ import annotations

import sys
from typing import TYPE_CHECKING, Any

import pytest
import upath

from llmling.config.models import (
    CallableResource,
    CLIResource,
    PathResource,
    Resource,
    SourceResource,
    TextResource,
)
from llmling.core import exceptions
from llmling.core.typedefs import ProcessingStep
from llmling.processors.base import ProcessorConfig
from llmling.processors.registry import ProcessorRegistry
from llmling.resources.loaders import (
    CallableResourceLoader,
    CLIResourceLoader,
    PathResourceLoader,
    SourceResourceLoader,
    TextResourceLoader,
)
from llmling.resources.models import LoadedResource
from tests.test_processors import REVERSED_TEXT


if TYPE_CHECKING:
    from pathlib import Path

    from llmling.resources.base import ResourceLoader

# Constants for test data
SAMPLE_TEXT = "Hello, World!"
TIMEOUT_SECONDS = 1
LARGE_TEXT = "A" * 1000
INVALID_MODULE = "does_not_exist.module"
INVALID_FUNCTION = "invalid_function"
TEST_FILE_CONTENT = "Test file content"
TEST_URL = "https://example.com/test.txt"
TEST_URL_CONTENT = "Test URL content"
GIT_HELP_COMMAND = "git --help"
LONG_RUNNING_COMMAND = "sleep 10"
ECHO_COMMAND = "echo test" if sys.platform == "win32" else ["echo", "test"]
SLEEP_COMMAND = "timeout 2" if sys.platform == "win32" else ["sleep", "2"]


# Test helpers
async def async_function(**kwargs: Any) -> str:
    """Test async function for callable loader."""
    return f"Async result with {kwargs}"


def sync_function(**kwargs: Any) -> str:
    """Test sync function for callable loader."""
    return f"Sync result with {kwargs}"


def failing_function(**kwargs: Any) -> str:
    """Test function that raises an exception."""
    msg = "Test error"
    raise ValueError(msg)


def reverse_text(text: str) -> str:
    """Helper function to reverse text."""
    return text[::-1]


# Fixtures


@pytest.fixture
def tmp_file(tmp_path: Path) -> Path:
    """Create a temporary test file."""
    test_file = tmp_path / "test.txt"
    test_file.write_text(TEST_FILE_CONTENT)
    return test_file


# Text Loader Tests
@pytest.mark.asyncio
async def test_text_loader_basic() -> None:
    """Test basic text loading functionality."""
    context = TextResource(content=SAMPLE_TEXT, description="Test text")
    loader = TextResourceLoader()
    result = await loader.load(context, ProcessorRegistry())

    assert isinstance(result, LoadedResource)
    assert result.content == SAMPLE_TEXT
    assert result.metadata.extra["type"] == "text"


@pytest.mark.asyncio
async def test_text_loader_with_processors(processor_registry: ProcessorRegistry) -> None:
    """Test text loading with processors."""
    await processor_registry.startup()
    try:
        path = "llmling.testing.processors.reverse_text"
        cfg = ProcessorConfig(type="function", import_path=path)
        processor_registry.register("reverse", cfg)
        steps = [ProcessingStep(name="reverse")]
        context = TextResource(content=SAMPLE_TEXT, description="test", processors=steps)
        loader = TextResourceLoader()
        result = await loader.load(context, processor_registry)
        assert result.content == REVERSED_TEXT
    finally:
        await processor_registry.shutdown()


# Path Loader Tests
@pytest.mark.asyncio
async def test_path_loader_file(tmp_file: Path) -> None:
    """Test loading from a file."""
    context = PathResource(path=str(tmp_file), description="Test file")
    loader = PathResourceLoader()
    result = await loader.load(context, ProcessorRegistry())

    assert result.content == TEST_FILE_CONTENT
    assert result.metadata.extra["type"] == "path"
    assert result.metadata.extra["path"] == str(tmp_file)


@pytest.mark.asyncio
async def test_path_loader_with_file_protocol(tmp_path: Path) -> None:
    """Test loading from a path with file:// protocol."""
    # Create a test file
    test_file = tmp_path / "test.txt"
    test_file.write_text(TEST_FILE_CONTENT)

    # Use UPath to create the proper file:// URL
    path = upath.UPath(test_file)
    file_url = str(path.as_uri())  # This will create the correct file:// URL format

    context = PathResource(path=file_url, description="Test file URL")

    loader = PathResourceLoader()
    result = await loader.load(context, ProcessorRegistry())

    assert result.content == TEST_FILE_CONTENT
    assert result.metadata.extra["path"] == file_url
    assert result.metadata.extra["scheme"] == "file"
    assert result.metadata.size == len(TEST_FILE_CONTENT)


@pytest.mark.asyncio
async def test_path_loader_error() -> None:
    """Test loading from a non-existent path."""
    context = PathResource(path="/nonexistent/file.txt", description="Test missing file")
    loader = PathResourceLoader()

    with pytest.raises(exceptions.LoaderError):
        await loader.load(context, ProcessorRegistry())


# CLI Loader Tests
@pytest.mark.asyncio
async def test_cli_loader_basic() -> None:
    """Test basic CLI command execution."""
    is_shell = sys.platform == "win32"
    context = CLIResource(
        command=ECHO_COMMAND, description="Test command", shell=is_shell
    )
    loader = CLIResourceLoader()
    result = await loader.load(context, ProcessorRegistry())

    assert "test" in result.content.strip()
    assert result.metadata.extra["exit_code"] == 0


@pytest.mark.asyncio
async def test_cli_loader_timeout() -> None:
    """Test CLI command timeout."""
    context = CLIResource(command=SLEEP_COMMAND, timeout=0.1, description="test")
    loader = CLIResourceLoader()

    with pytest.raises(exceptions.LoaderError):
        await loader.load(context, ProcessorRegistry())


# Source Loader Tests
@pytest.mark.asyncio
async def test_source_loader_basic() -> None:
    """Test basic source code loading."""
    path = "llmling.resources.loaders.text"
    context = SourceResource(import_path=path, description="Test source")
    loader = SourceResourceLoader()
    result = await loader.load(context, ProcessorRegistry())

    assert "class TextResourceLoader" in result.content
    assert result.metadata.extra["import_path"] == context.import_path


@pytest.mark.asyncio
async def test_source_loader_invalid_module() -> None:
    """Test loading from non-existent module."""
    context = SourceResource(
        import_path=INVALID_MODULE, description="Test invalid module"
    )
    loader = SourceResourceLoader()

    with pytest.raises(exceptions.LoaderError):
        await loader.load(context, ProcessorRegistry())


# Callable Loader Tests
@pytest.mark.asyncio
async def test_callable_loader_sync() -> None:
    """Test loading from synchronous callable."""
    context = CallableResource(
        import_path=f"{__name__}.sync_function",
        description="Test sync callable",
        keyword_args={"test": "value"},
    )
    loader = CallableResourceLoader()
    result = await loader.load(context, ProcessorRegistry())

    assert "Sync result with" in result.content
    assert result.metadata.extra["import_path"] == context.import_path


@pytest.mark.asyncio
async def test_callable_loader_async() -> None:
    """Test loading from asynchronous callable."""
    context = CallableResource(
        import_path=f"{__name__}.async_function",
        description="Test async callable",
        keyword_args={"test": "value"},
    )
    loader = CallableResourceLoader()
    result = await loader.load(context, ProcessorRegistry())

    assert "Async result with" in result.content
    assert result.metadata.extra["import_path"] == context.import_path


# Integration Tests
@pytest.mark.asyncio
async def test_all_loaders_with_processors(
    processor_registry: ProcessorRegistry,
    tmp_file: Path,
) -> None:
    """Test all loaders with processor chain."""
    cfg = ProcessorConfig(type="function", import_path="reprlib.repr")
    processor_registry.register("upper", cfg)
    cfg = ProcessorConfig(type="function", import_path=f"{__name__}.reverse_text")
    processor_registry.register("reverse", cfg)
    processors = [ProcessingStep(name="upper"), ProcessingStep(name="reverse")]

    resources: list[Resource] = [
        TextResource(content=SAMPLE_TEXT, description="Test text", processors=processors),
        PathResource(path=str(tmp_file), description="Test file", processors=processors),
        CLIResource(
            command=ECHO_COMMAND,
            description="Test command",
            shell=sys.platform == "win32",
            processors=processors,
        ),
    ]

    loaders: dict[str, ResourceLoader[Any]] = {
        "text": TextResourceLoader(),
        "path": PathResourceLoader(),
        "cli": CLIResourceLoader(),
    }

    for context in resources:
        loader = loaders[context.resource_type]
        result = await loader.load(context, processor_registry)
        assert isinstance(result, LoadedResource)
        assert result.content
        assert result.content.startswith("'")


if __name__ == "__main__":
    pytest.main(["-vv", __file__])



"""Test server lifecycle using different approaches."""

from __future__ import annotations

import asyncio
import contextlib
import json
import sys
from typing import TYPE_CHECKING, Any

import logfire
import pytest


if TYPE_CHECKING:
    from llmling.server import LLMLingServer
    from llmling.testing.testclient import MCPInProcSession


# Initialize logfire to avoid warnings
logfire.configure()


@pytest.mark.asyncio
async def test_server_lifecycle_handshake_client(client: MCPInProcSession) -> None:
    """Test server lifecycle using MCPInProcSession."""
    try:
        await client.start()
        await asyncio.sleep(0.5)

        # Initialize connection
        init_response = await client.send_request(
            "initialize",
            {
                "protocolVersion": "2024-11-05",
                "capabilities": {},
                "clientInfo": {"name": "test-client", "version": "1.0"},
            },
        )
        assert isinstance(init_response, dict)
        assert "serverInfo" in init_response
        assert init_response["serverInfo"]["name"] == "llmling-server"

        # Send initialized notification
        await client.send_notification("notifications/initialized", {})

        # Test tool listing
        tools_response = await client.send_request("tools/list")
        assert isinstance(tools_response, dict)
        assert "tools" in tools_response
        assert isinstance(tools_response["tools"], list)
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_server_lifecycle_test_session(
    running_server: tuple[LLMLingServer, tuple[Any, Any]],
) -> None:
    """Test server lifecycle using test session."""
    server, (client_read, client_write) = running_server

    # Create proper MCP message
    from mcp.types import JSONRPCMessage, JSONRPCRequest

    init_request = JSONRPCMessage(
        JSONRPCRequest(
            jsonrpc="2.0",
            id=1,
            method="initialize",
            params={
                "protocolVersion": "2024-11-05",
                "capabilities": {},
                "clientInfo": {"name": "test-client", "version": "1.0"},
            },
        )
    )

    await client_write.send(init_request)
    response = await client_read.receive()
    assert "result" in response.root.model_dump()
    assert "serverInfo" in response.root.result

    # Send initialized notification
    from mcp.types import JSONRPCNotification

    notification = JSONRPCMessage(
        JSONRPCNotification(
            jsonrpc="2.0",
            method="notifications/initialized",
            params={},
        )
    )
    await client_write.send(notification)

    # Test tools list
    tools_request = JSONRPCMessage(
        JSONRPCRequest(
            jsonrpc="2.0",
            id=2,
            method="tools/list",
            params={},
        )
    )
    await client_write.send(tools_request)
    tools_response = await client_read.receive()
    assert "tools" in tools_response.root.result


@pytest.mark.asyncio
async def test_server_lifecycle_direct(server: LLMLingServer) -> None:
    """Test server lifecycle using direct method calls."""
    try:
        # Start registries
        await server.processor_registry.startup()
        await server.tool_registry.startup()

        # Test registry contents directly
        tools = list(server.tool_registry.values())
        assert isinstance(tools, list)
        assert len(tools) > 0  # Should have our test tools
    finally:
        await server.shutdown()


@pytest.mark.asyncio
async def test_server_lifecycle_subprocess() -> None:
    """Test server lifecycle using raw subprocess."""
    process = await asyncio.create_subprocess_exec(
        sys.executable,
        "-m",
        "llmling.server",
        stdin=asyncio.subprocess.PIPE,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )

    # Start stderr reader task
    async def read_stderr():
        assert process.stderr
        while True:
            line = await process.stderr.readline()
            if not line:
                break
            print(f"Server stderr: {line.decode().strip()}")

    stderr_task = asyncio.create_task(read_stderr())

    try:
        assert process.stdin
        assert process.stdout
        await asyncio.sleep(0.5)  # Give server time to start

        # Send initialize request
        request = {
            "jsonrpc": "2.0",
            "id": 1,
            "method": "initialize",
            "params": {
                "protocolVersion": "2024-11-05",
                "capabilities": {},
                "clientInfo": {"name": "test-client", "version": "1.0"},
            },
        }
        process.stdin.write(json.dumps(request).encode() + b"\n")
        await process.stdin.drain()

        # Read until we get a valid JSON response
        while True:
            response = await process.stdout.readline()
            if not response:
                msg = "No response from server"
                raise RuntimeError(msg)

            try:
                _result = json.loads(response.decode())
                break  # Valid JSON found
            except json.JSONDecodeError:
                continue  # Skip non-JSON lines

        # Rest of the test...

    finally:
        stderr_task.cancel()
        with contextlib.suppress(asyncio.CancelledError):
            await stderr_task


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Callable context loader."""

from __future__ import annotations

import asyncio
import importlib
from typing import TYPE_CHECKING, Any, TypeGuard


if TYPE_CHECKING:
    from collections.abc import Awaitable, Callable


def is_async_callable(obj: Any) -> TypeGuard[Callable[..., Awaitable[Any]]]:
    """Check if an object is an async callable."""
    return asyncio.iscoroutinefunction(obj)


def import_callable(import_path: str) -> Callable[..., Any]:
    """Import a callable from an import path.

    Args:
        import_path: Dot-separated path to callable (e.g., "module.submodule.func")

    Returns:
        The imported callable

    Raises:
        ValueError: If import fails or object is not callable
    """
    try:
        module_path, callable_name = import_path.rsplit(".", 1)
        module = importlib.import_module(module_path)
        callable_obj = getattr(module, callable_name)

        if not callable(callable_obj):
            msg = f"Imported object {import_path} is not callable"
            raise ValueError(msg)  # noqa: TRY004
    except ImportError as exc:
        msg = f"Could not import callable: {import_path}"
        raise ValueError(msg) from exc
    except AttributeError as exc:
        msg = f"Could not find callable {import_path!r} "
        raise ValueError(msg) from exc
    else:
        return callable_obj


async def execute_callable(import_path: str, **kwargs: Any) -> Any:
    """Execute a callable and return its result.

    Args:
        import_path: Dot-separated path to callable
        **kwargs: Arguments to pass to the callable

    Returns:
        Result of the callable execution

    Raises:
        ValueError: If import or execution fails
    """
    try:
        callable_obj = import_callable(import_path)

        # Execute the callable
        if is_async_callable(callable_obj):
            result = await callable_obj(**kwargs)
        else:
            result = callable_obj(**kwargs)
    except Exception as exc:
        msg = f"Error executing callable {import_path}: {exc}"
        raise ValueError(msg) from exc
    else:
        return result



# llmling/core/registry.py
from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Iterator, MutableMapping
from typing import TYPE_CHECKING, Any, TypeVar


if TYPE_CHECKING:
    from collections.abc import Sequence

    from llmling.core import exceptions
    from llmling.core.events import RegistryEvents


TItem = TypeVar("TItem")
TKey = TypeVar("TKey", str, int)


class BaseRegistry[TKey, TItem](MutableMapping[TKey, TItem], ABC):
    """Base class for component registries."""

    def __init__(self) -> None:
        """Initialize an empty registry."""
        self._items: dict[TKey, TItem] = {}
        self._initialized = False
        self._configs: dict[TKey, Any] = {}
        self._observers: set[RegistryEvents[TKey, TItem]] = set()

    @property
    def is_empty(self) -> bool:
        """Check if registry has any items."""
        return not bool(self._items)

    def has_item(self, key: TKey) -> bool:
        """Check if an item is registered."""
        return key in self._items

    def register(self, key: TKey, item: TItem | Any, replace: bool = False) -> None:
        """Register an item."""
        if key in self._items and not replace:
            msg = f"Item already registered: {key}"
            raise self._error_class(msg)

        validated_item = self._validate_item(item)
        is_new = key not in self._items
        self._items[key] = validated_item

        if is_new:
            self._notify_item_added(key, validated_item)
        else:
            self._notify_item_modified(key, validated_item)

    def get(self, key: TKey) -> TItem:  # type: ignore
        """Get an item by key."""
        return self[key]

    def list_items(self) -> Sequence[TKey]:
        """List all registered item keys."""
        return list(self._items.keys())

    def reset(self) -> None:
        """Reset registry to initial state."""
        self._items.clear()
        self._configs.clear()
        self._notify_reset()
        self._initialized = False

    async def startup(self) -> None:
        """Initialize all registered items."""
        if self._initialized:
            return

        try:
            for item in self._items.values():
                await self._initialize_item(item)
            self._initialized = True
        except Exception as exc:
            await self.shutdown()
            msg = f"Registry startup failed: {exc}"
            raise self._error_class(msg) from exc

    async def shutdown(self) -> None:
        """Cleanup all registered items."""
        if not self._initialized:
            return

        errors: list[tuple[TKey, Exception]] = []

        for key, item in self._items.items():
            try:
                await self._cleanup_item(item)
            except Exception as exc:  # noqa: BLE001
                errors.append((key, exc))

        self._initialized = False

        if errors:
            error_msgs = [f"{key}: {exc}" for key, exc in errors]
            msg = f"Errors during shutdown: {', '.join(error_msgs)}"
            raise self._error_class(msg)

    @property
    @abstractmethod
    def _error_class(self) -> type[exceptions.LLMLingError]:
        """Error class to use for this registry."""

    @abstractmethod
    def _validate_item(self, item: Any) -> TItem:
        """Validate and possibly transform item before registration."""

    async def _initialize_item(self, item: TItem) -> None:
        """Initialize an item during startup."""
        if hasattr(item, "startup") and callable(item.startup):
            await item.startup()

    async def _cleanup_item(self, item: TItem) -> None:
        """Clean up an item during shutdown."""
        if hasattr(item, "shutdown") and callable(item.shutdown):
            await item.shutdown()

    # Implementing MutableMapping methods
    def __getitem__(self, key: TKey) -> TItem:
        try:
            return self._items[key]
        except KeyError as exc:
            msg = f"Item not found: {key}"
            raise self._error_class(msg) from exc

    def __setitem__(self, key: TKey, value: Any) -> None:
        """Support dict-style assignment."""
        self.register(key, value)

    def __contains__(self, key: object) -> bool:
        """Support 'in' operator without raising exceptions."""
        return key in self._items

    def __delitem__(self, key: TKey) -> None:
        if key in self._items:
            item = self._items[key]
            del self._items[key]
            self._notify_item_removed(key, item)
        else:
            msg = f"Item not found: {key}"
            raise self._error_class(msg)

    def __iter__(self) -> Iterator[TKey]:
        return iter(self._items)

    async def __aiter__(self):
        """Async iterate over items, ensuring they're initialized."""
        if not self._initialized:
            await self.startup()
        for key, item in self._items.items():
            yield key, item

    def __len__(self) -> int:
        return len(self._items)

    def add_observer(self, observer: RegistryEvents[TKey, TItem]) -> None:
        """Add an observer for registry changes."""
        self._observers.add(observer)

    def remove_observer(self, observer: RegistryEvents[TKey, TItem]) -> None:
        """Remove a registered observer."""
        self._observers.discard(observer)

    def _notify_item_added(self, key: TKey, item: TItem) -> None:
        """Notify observers about item addition."""
        for obs in self._observers:
            if obs.on_item_added:
                obs.on_item_added(key, item)

    def _notify_item_removed(self, key: TKey, item: TItem) -> None:
        """Notify observers about item removal."""
        for obs in self._observers:
            if obs.on_item_removed:
                obs.on_item_removed(key, item)

    def _notify_item_modified(self, key: TKey, item: TItem) -> None:
        """Notify observers about item modification."""
        for obs in self._observers:
            if obs.on_item_modified:
                obs.on_item_modified(key, item)

    def _notify_list_changed(self) -> None:
        """Notify observers about list changes."""
        for obs in self._observers:
            if obs.on_list_changed:
                obs.on_list_changed()

    def _notify_reset(self) -> None:
        """Notify observers about registry reset."""
        for obs in self._observers:
            if obs.on_reset:
                obs.on_reset()



from __future__ import annotations

from contextlib import asynccontextmanager
from typing import TYPE_CHECKING, TypeVar, cast

import anyio
from anyio.streams.memory import MemoryObjectReceiveStream, MemoryObjectSendStream
from mcp.server.models import InitializationOptions
from mcp.server.session import ServerSession
from mcp.types import (
    JSONRPCMessage,
    LoggingCapability,
    PromptsCapability,
    ResourcesCapability,
    ServerCapabilities,
    ToolsCapability,
)


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator


T_Stream = TypeVar("T_Stream")


class TestStreamPair[T_Stream]:
    """Test stream implementation."""

    def __init__(self) -> None:
        """Initialize stream pair."""
        # Create streams
        streams = anyio.create_memory_object_stream[T_Stream](max_buffer_size=100)
        # Use explicit casting to help mypy understand the types
        self.receive_stream = cast(
            MemoryObjectReceiveStream[T_Stream | Exception], streams[0]
        )
        self.send_stream = cast(MemoryObjectSendStream[T_Stream], streams[1])


@asynccontextmanager
async def create_test_server_session() -> AsyncGenerator[
    tuple[
        ServerSession,
        tuple[
            MemoryObjectReceiveStream[JSONRPCMessage],  # Client read stream
            MemoryObjectSendStream[JSONRPCMessage],  # Client write stream
        ],
    ],
    None,
]:
    """Create a test server session with streams."""
    # Create the streams
    # First stream: Client write -> Server read (with Exception support)
    send_exc, recv_exc = anyio.create_memory_object_stream[JSONRPCMessage | Exception]()

    # Second stream: Server write -> Client read
    send_msg, recv_msg = anyio.create_memory_object_stream[JSONRPCMessage]()

    # Create server session
    session = ServerSession(
        read_stream=recv_exc,  # Server reads (receives) from client's send stream
        write_stream=send_msg,  # Server writes (sends) to client's receive stream
        init_options=InitializationOptions(
            server_name="test-server",
            server_version="1.0.0",
            capabilities=ServerCapabilities(
                resources=ResourcesCapability(
                    subscribe=True,
                    listChanged=True,
                ),
                prompts=PromptsCapability(
                    listChanged=True,
                ),
                tools=ToolsCapability(
                    listChanged=True,
                ),
                logging=LoggingCapability(),
            ),
        ),
    )

    async with session:
        # Return the client's ends of the streams
        yield session, (recv_msg, send_exc)



"""Resource loading functionality."""

from llmling.resources.base import ResourceLoader
from llmling.resources.loaders import (
    CallableResourceLoader,
    CLIResourceLoader,
    PathResourceLoader,
    SourceResourceLoader,
    TextResourceLoader,
)
from llmling.resources.loaders.registry import ResourceLoaderRegistry
from llmling.resources.registry import ResourceRegistry
from llmling.resources.models import LoadedResource
from llmling.resources.loaders.image import ImageResourceLoader

# Create and populate the default registry
default_registry = ResourceLoaderRegistry()
default_registry["image"] = ImageResourceLoader
default_registry["path"] = PathResourceLoader
default_registry["text"] = TextResourceLoader
default_registry["cli"] = CLIResourceLoader
default_registry["source"] = SourceResourceLoader
default_registry["callable"] = CallableResourceLoader

__all__ = [
    "CLIResourceLoader",
    "CallableResourceLoader",
    "ImageResourceLoader",
    "LoadedResource",
    "PathResourceLoader",
    "ResourceLoader",
    "ResourceLoaderRegistry",
    "ResourceRegistry",
    "SourceResourceLoader",
    "TextResourceLoader",
    "default_registry",
]



"""Resource loader implementations."""

from llmling.resources.loaders.callable import CallableResourceLoader
from llmling.resources.loaders.cli import CLIResourceLoader
from llmling.resources.loaders.path import PathResourceLoader
from llmling.resources.loaders.source import SourceResourceLoader
from llmling.resources.loaders.text import TextResourceLoader

__all__ = [
    "CLIResourceLoader",
    "CallableResourceLoader",
    "PathResourceLoader",
    "SourceResourceLoader",
    "TextResourceLoader",
]



from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

from logfire import instrument

from llmling.core import exceptions
from llmling.core.log import get_logger
from llmling.core.typedefs import MessageContent
from llmling.prompts.models import PromptMessage, PromptResult, ResolvedContent


if TYPE_CHECKING:
    from llmling.processors.registry import ProcessorRegistry
    from llmling.prompts.models import Prompt
    from llmling.resources import ResourceLoaderRegistry

logger = get_logger(__name__)


@instrument("Resolving resources")
async def resolve_resources(
    message: PromptMessage,
    resource_registry: ResourceLoaderRegistry,
    processor_registry: ProcessorRegistry,
) -> PromptMessage:
    """Resolve resources in a message.

    Args:
        message: Message to resolve
        resource_registry: Registry for loading resources
        processor_registry: Registry for content processors

    Returns:
        Message with resolved resources

    Raises:
        ResourceResolutionError: If resource resolution fails
        ProcessorError: If content processing fails
    """
    if not message.needs_resolution():
        return message

    now = datetime.now()
    resolved: list[ResolvedContent] = []

    for content in message.get_content_items():
        if content.type != "resource":
            resolved.append(ResolvedContent(original=content))
            continue

        try:
            # Find appropriate loader
            loader = resource_registry.find_loader_for_uri(content.content)
            from llmling.resources.base import ResourceLoader

            if not isinstance(loader, ResourceLoader):
                msg = f"Invalid loader type for {content.content}"
                raise exceptions.ResourceResolutionError(msg)  # noqa: TRY301

            # Load resource
            try:
                resource = await loader.load(loader.context, processor_registry)
            except Exception as exc:
                msg = f"Failed to load resource {content.content}: {exc}"
                raise exceptions.ResourceResolutionError(msg) from exc

            resolved.append(
                ResolvedContent(
                    original=content,
                    resolved=resource,
                    resolved_at=now,
                )
            )
        except exceptions.ResourceResolutionError:
            # Re-raise specific resource errors
            raise
        except Exception as exc:
            msg = f"Unexpected error resolving {content.content}: {exc}"
            raise exceptions.ResourceResolutionError(msg) from exc

    return PromptMessage(
        role=message.role,
        content=message.content,
        resolved_content=resolved,
    )


@instrument("Rendering prompt")
async def render_prompt(
    prompt: Prompt,
    arguments: dict[str, Any],
    *,
    resource_registry: ResourceLoaderRegistry | None = None,
    processor_registry: ProcessorRegistry | None = None,
) -> PromptResult:
    """Render a prompt template with arguments.

    Args:
        prompt: Prompt to render
        arguments: Arguments for template
        resource_registry: Optional registry for resolving resources
        processor_registry: Optional registry for content processors

    Returns:
        Rendered prompt result

    Raises:
        ProcessorError: If rendering fails
    """
    try:
        # Validate arguments first
        prompt.validate_arguments(arguments)

        rendered_messages: list[PromptMessage] = []
        now = datetime.now()

        for message in prompt.messages:
            # Handle string content
            if isinstance(message.content, str):
                text = message.content.format(**arguments)
                msg = PromptMessage(role=message.role, content=text)
                rendered_messages.append(msg)
                continue

            # Handle complex content
            contents = message.get_content_items()
            rendered_contents: list[MessageContent] = []

            for content in contents:
                if content.type == "text":
                    # Format text content
                    text = content.content.format(**arguments)
                    rendered_contents.append(MessageContent.text(text))
                else:
                    # Keep other content types as-is
                    rendered_contents.append(content)

            rendered_message = PromptMessage(
                role=message.role,
                content=rendered_contents,
            )

            # Resolve resources if registries provided
            if (
                resource_registry
                and processor_registry
                and rendered_message.needs_resolution()
            ):
                rendered_message = await resolve_resources(
                    rendered_message,
                    resource_registry,
                    processor_registry,
                )

            rendered_messages.append(rendered_message)

        return PromptResult(
            messages=rendered_messages,
            metadata={"prompt_name": prompt.name, "arguments": arguments},
            resolved_at=now if resource_registry else None,
        )

    except KeyError as exc:
        error_msg = f"Missing argument in template: {exc}"
        raise exceptions.ProcessorError(error_msg) from exc
    except ValueError as exc:
        error_msg = f"Invalid arguments: {exc}"
        raise exceptions.ProcessorError(error_msg) from exc
    except Exception as exc:
        error_msg = f"Failed to render prompt: {exc}"
        raise exceptions.ProcessorError(error_msg) from exc



"""Tests for the processor system."""

from __future__ import annotations

import asyncio

import pytest

from llmling.core import exceptions
from llmling.core.typedefs import ProcessingStep
from llmling.processors.base import ProcessorConfig
from llmling.processors.implementations.function import FunctionProcessor
from llmling.processors.implementations.template import TemplateProcessor
from llmling.processors.registry import ProcessorRegistry
from llmling.resources.models import ProcessingContext


# Test data
SAMPLE_TEXT = "Hello, World!"
REVERSED_TEXT = SAMPLE_TEXT[::-1]
UPPER_TEXT = SAMPLE_TEXT.upper()
REVERSE_IMPORT = "llmling.testing.processors.reverse_text"
UPPERCASE_IMPORT = "llmling.testing.processors.uppercase_text"
APPEND_IMPORT = "llmling.testing.processors.append_text"
FAILING_IMPORT = "llmling.testing.processors.failing_processor"


# Test helpers
def sync_reverse(text: str) -> str:
    """Test helper to reverse text."""
    return text[::-1]


async def async_reverse(text: str) -> str:
    """Test helper to reverse text asynchronously."""
    await asyncio.sleep(0.1)
    return text[::-1]


async def failing_processor(text: str) -> str:
    """Test helper that fails."""
    msg = "Test failure"
    raise ValueError(msg)


# Test fixtures
@pytest.fixture
def function_config() -> ProcessorConfig:
    """Create a test function processor config."""
    return ProcessorConfig(type="function", import_path=REVERSE_IMPORT)


@pytest.fixture
def template_config() -> ProcessorConfig:
    """Create a test template processor config."""
    return ProcessorConfig(
        type="template",
        template="Processed: {{ content }}",
    )


@pytest.fixture
def registry() -> ProcessorRegistry:
    """Create and initialize a test processor registry."""
    return ProcessorRegistry()


@pytest.mark.asyncio
async def test_processor_pipeline(registry: ProcessorRegistry) -> None:
    """Test complete processor pipeline."""
    # Register processors
    cfg = ProcessorConfig(type="function", import_path=UPPERCASE_IMPORT)
    cfg_2 = ProcessorConfig(type="function", import_path=APPEND_IMPORT)
    registry.register("upper", cfg)
    registry.register("append", cfg_2)

    # Define processing steps
    steps = [
        ProcessingStep(name="upper"),
        ProcessingStep(name="append", kwargs={"suffix": "!!!"}),
    ]

    # Process text
    try:
        await registry.startup()
        result = await registry.process("hello", steps)

        assert result.content == "HELLO!!!"
        assert result.original_content == "hello"
        assert "function" in result.metadata
    finally:
        await registry.shutdown()


@pytest.mark.asyncio
async def test_function_processor() -> None:
    """Test function processor execution."""
    config = ProcessorConfig(type="function", import_path=REVERSE_IMPORT)

    processor = FunctionProcessor(config)

    try:
        await processor.startup()
        ctx = ProcessingContext(original_content=SAMPLE_TEXT, current_content=SAMPLE_TEXT)
        result = await processor.process(ctx)

        assert result.content == REVERSED_TEXT
        assert result.metadata["function"] == REVERSE_IMPORT
        assert not result.metadata["is_async"]
    finally:
        await processor.shutdown()


@pytest.fixture
def initialized_registry(registry: ProcessorRegistry) -> ProcessorRegistry:
    """Create and initialize a test processor registry."""
    # Register configurations
    cfg = ProcessorConfig(type="function", import_path=REVERSE_IMPORT)
    cfg_2 = ProcessorConfig(type="function", import_path=REVERSE_IMPORT)
    registry.register("reverse", cfg)
    registry.register("reverse1", cfg_2)
    return registry


@pytest.fixture
def processor_registry(registry: ProcessorRegistry) -> ProcessorRegistry:
    """Create a processor registry for complex tests."""
    return registry


# Base processor tests
@pytest.mark.asyncio
async def test_processor_lifecycle(function_config: ProcessorConfig) -> None:
    """Test processor startup and shutdown."""
    processor = FunctionProcessor(function_config)

    await processor.startup()
    assert processor.func is not None

    ctx = ProcessingContext(original_content=SAMPLE_TEXT, current_content=SAMPLE_TEXT)
    result = await processor.process(ctx)
    assert result.content == REVERSED_TEXT

    await processor.shutdown()


@pytest.mark.asyncio
async def test_processor_validation(function_config: ProcessorConfig) -> None:
    """Test processor result validation."""
    function_config.validate_output = True
    processor = FunctionProcessor(function_config)

    await processor.startup()
    ctx = ProcessingContext(original_content=SAMPLE_TEXT, current_content=SAMPLE_TEXT)
    result = await processor.process(ctx)
    assert result.content == REVERSED_TEXT


# Function processor tests
@pytest.mark.asyncio
async def test_function_processor_sync() -> None:
    """Test synchronous function processor."""
    config = ProcessorConfig(type="function", import_path=REVERSE_IMPORT)
    processor = FunctionProcessor(config)

    await processor.startup()
    ctx = ProcessingContext(original_content=SAMPLE_TEXT, current_content=SAMPLE_TEXT)
    result = await processor.process(ctx)

    assert result.content == REVERSED_TEXT
    assert result.metadata["function"] == REVERSE_IMPORT
    assert not result.metadata["is_async"]


@pytest.mark.asyncio
async def test_function_processor_async() -> None:
    """Test asynchronous function processor."""
    config = ProcessorConfig(
        type="function",
        name="async_reverse",
        import_path="llmling.testing.processors.async_reverse_text",
    )
    processor = FunctionProcessor(config)

    await processor.startup()
    ctx = ProcessingContext(original_content=SAMPLE_TEXT, current_content=SAMPLE_TEXT)
    result = await processor.process(ctx)

    assert result.content == REVERSED_TEXT
    assert result.metadata["function"] == "llmling.testing.processors.async_reverse_text"
    assert result.metadata["is_async"]


@pytest.mark.asyncio
async def test_function_processor_error() -> None:
    """Test function processor error handling."""
    config = ProcessorConfig(
        type="function",
        name="failing_processor",
        import_path=FAILING_IMPORT,
    )
    processor = FunctionProcessor(config)

    await processor.startup()
    ctx = ProcessingContext(original_content=SAMPLE_TEXT, current_content=SAMPLE_TEXT)
    with pytest.raises(exceptions.ProcessorError, match="Function execution failed"):
        await processor.process(ctx)


# Template processor tests
@pytest.mark.asyncio
async def test_template_processor_basic(template_config: ProcessorConfig) -> None:
    """Test basic template processing."""
    processor = TemplateProcessor(template_config)

    await processor.startup()
    context = ProcessingContext(
        original_content=SAMPLE_TEXT,
        current_content=SAMPLE_TEXT,
        kwargs={"extra": "value"},
    )

    result = await processor.process(context)

    assert result.content == f"Processed: {SAMPLE_TEXT}"
    assert "content" in result.metadata["template_vars"]
    assert "extra" in result.metadata["template_vars"]


@pytest.mark.asyncio
async def test_template_processor_error() -> None:
    """Test template processor error handling."""
    # Force a template error
    config = ProcessorConfig(type="template", template="{{ undefined_var + 123 }}")
    processor = TemplateProcessor(config)

    await processor.startup()
    ctx = ProcessingContext(original_content=SAMPLE_TEXT, current_content=SAMPLE_TEXT)
    with pytest.raises(exceptions.ProcessorError):
        await processor.process(ctx)


# Registry tests
@pytest.mark.asyncio
async def test_registry_lifecycle(registry: ProcessorRegistry) -> None:
    """Test registry startup and shutdown."""
    # Create config first, with all required fields
    config = ProcessorConfig(
        type="function",
        import_path=f"{__name__}.sync_reverse",
        # Don't set name directly, let the validator handle it
    )

    registry.register("reverse", config)
    await registry.startup()
    assert registry._initialized
    await registry.shutdown()
    assert not registry._initialized


@pytest.mark.asyncio
async def test_registry_sequential_processing(
    initialized_registry: ProcessorRegistry,
) -> None:
    """Test sequential processing."""
    await initialized_registry.startup()
    try:
        steps = [ProcessingStep(name="reverse")]
        result = await initialized_registry.process("hello", steps)
        assert result.content == "olleh"
    finally:
        await initialized_registry.shutdown()


@pytest.mark.asyncio
async def test_registry_parallel_processing(registry: ProcessorRegistry) -> None:
    """Test parallel processing."""
    # Register processors first
    cfg = ProcessorConfig(type="function", import_path=REVERSE_IMPORT)
    cfg_2 = ProcessorConfig(type="function", import_path=REVERSE_IMPORT)
    registry.register("reverse1", cfg)
    registry.register("reverse2", cfg_2)

    # Then start the registry
    await registry.startup()

    steps = [
        ProcessingStep(name="reverse1", parallel=True),
        ProcessingStep(name="reverse2", parallel=True),
    ]

    try:
        result = await registry.process(SAMPLE_TEXT, steps)
        assert REVERSED_TEXT in result.content
    finally:
        await registry.shutdown()


@pytest.mark.asyncio
async def test_registry_streaming(registry: ProcessorRegistry) -> None:
    p = "llmling.testing.processors.async_reverse_text"
    registry.register("reverse", ProcessorConfig(type="function", import_path=p))

    steps = [ProcessingStep(name="reverse")]
    results = [result async for result in registry.process_stream(SAMPLE_TEXT, steps)]
    assert len(results) == 1
    assert results[0].content == REVERSED_TEXT


@pytest.mark.asyncio
async def test_registry_optional_step(registry: ProcessorRegistry) -> None:
    cfg = ProcessorConfig(type="function", import_path=FAILING_IMPORT)
    registry.register("fail", cfg)
    cfg = ProcessorConfig(type="function", import_path=REVERSE_IMPORT)
    registry.register("reverse", cfg)

    steps = [
        ProcessingStep(name="fail", required=False),
        ProcessingStep(name="reverse"),
    ]

    result = await registry.process(SAMPLE_TEXT, steps)
    assert result.content == REVERSED_TEXT


@pytest.mark.asyncio
async def test_registry_error_handling(registry: ProcessorRegistry) -> None:
    """Test registry error handling."""
    p = f"{__name__}.failing_processor"
    cfg = ProcessorConfig(type="function", name="failing_processor", import_path=p)
    registry.register("fail", cfg)
    steps = [ProcessingStep(name="fail")]
    with pytest.raises(exceptions.ProcessorError):
        await registry.process(SAMPLE_TEXT, steps)


# Integration tests
@pytest.mark.asyncio
async def test_complex_processing_pipeline(processor_registry: ProcessorRegistry) -> None:
    """Test complex processing pipeline with sequential and parallel steps."""
    await processor_registry.startup()
    try:
        # Register processors
        cfg = ProcessorConfig(type="function", import_path=f"{__name__}.sync_reverse")
        processor_registry.register("reverse", cfg)
        cfg = ProcessorConfig(type="template", template="First: {{ content }}")
        processor_registry.register("template1", cfg)
        cfg = ProcessorConfig(type="template", template="Second: {{ content }}")
        processor_registry.register("template2", cfg)

        steps = [
            ProcessingStep(name="template1"),
            ProcessingStep(name="template2"),
            ProcessingStep(name="reverse"),
        ]

        result = await processor_registry.process("Hello, World!", steps)
        content = result.content

        # The content is reversed, so we need to reverse it back to check
        unreversed = content[::-1]
        assert any(marker in unreversed for marker in ["First:", "Second:"]), (
            f"Content (unreversed) does not contain expected markers: {unreversed}"
        )

        # Or we could check for the reversed markers
        assert any(marker in content for marker in [":tsriF", ":dnoceS"]), (
            f"Content does not contain reversed markers: {content}"
        )

    finally:
        await processor_registry.shutdown()


if __name__ == "__main__":
    pytest.main([__file__, "-vv"])



"""File system watching implementation."""

from __future__ import annotations

import asyncio
import os
import pathlib
import time
from typing import TYPE_CHECKING

import pathspec
import upath
from watchdog.events import FileSystemEvent, FileSystemEventHandler
from watchdog.observers import Observer

from llmling.core.log import get_logger
from llmling.resources.watching.utils import debounce


if TYPE_CHECKING:
    from watchdog.observers.api import BaseObserver

    from llmling.config.models import Resource
    from llmling.resources.registry import ResourceRegistry


logger = get_logger(__name__)


def is_watchable_path(path: str | os.PathLike[str]) -> bool:
    """Check if a path can be watched with watchdog.

    Args:
        path: Path to check

    Returns:
        True if path is local and can be watched
    """
    return upath.UPath(path).protocol in ("file", "")


class ResourceEventHandler(FileSystemEventHandler):
    """Handles file system events for a specific resource."""

    def __init__(
        self,
        resource_name: str,
        registry: ResourceRegistry,
        patterns: list[str],
        loop: asyncio.AbstractEventLoop,
    ) -> None:
        self.resource_name = resource_name
        self.registry = registry
        self.loop = loop
        self.spec = pathspec.PathSpec.from_lines(
            pathspec.patterns.GitWildMatchPattern,
            patterns,
        )
        self._debounced_notify = debounce(wait=0.1)(self._notify_change)
        self._last_notification = 0.0  # Track last notification time

    def on_any_event(self, event: FileSystemEvent) -> None:
        """Handle any file system event."""
        if event.is_directory:
            return

        path = (
            event.src_path.decode()
            if isinstance(event.src_path, bytes)
            else event.src_path
        )

        # Convert to relative path for matching
        path = os.path.relpath(path, start=pathlib.Path(path).parent)

        if self.spec.match_file(path):
            # Use loop directly for thread-safety
            self.loop.call_soon_threadsafe(self._notify_change)

    def _notify_change(self) -> None:
        """Notify registry of resource change."""
        try:
            now = time.time()
            # Debounce notifications
            if now - self._last_notification < 0.1:  # noqa: PLR2004
                return

            self.registry.invalidate(self.resource_name)
            self._last_notification = now

        except Exception:
            logger.exception("Failed to notify change for %s", self.resource_name)


class ResourceWatcher:
    """Manages file system watching for resources."""

    def __init__(self, registry: ResourceRegistry) -> None:
        """Initialize watcher.

        Args:
            registry: Registry to notify of changes
        """
        self.registry = registry
        self.observer: BaseObserver | None = None
        self.handlers: dict[str, ResourceEventHandler] = {}
        # Track watch directories to avoid duplicates
        self._watched_paths: set[str] = set()
        self.loop = asyncio.get_event_loop()

    async def start(self) -> None:
        """Start the file system observer."""
        if self.observer:
            return

        try:
            self.observer = Observer()
            self.observer.start()
            logger.info("File system watcher started")
        except Exception:
            self.observer = None
            logger.exception("Failed to start file system watcher")

    async def stop(self) -> None:
        """Stop the file system observer."""
        if not self.observer:
            return

        try:
            self.observer.stop()
            self.observer.join(timeout=2.0)
            self.observer = None
            self.handlers.clear()
            self._watched_paths.clear()
            logger.info("File system watcher stopped")
        except Exception:
            logger.exception("Error stopping file system watcher")

    def add_watch(self, name: str, resource: Resource) -> None:
        """Add a watch for a resource.

        Args:
            name: Resource name
            resource: Resource to watch
        """
        if not self.observer:
            return

        try:
            # Create handler with loop
            handler = ResourceEventHandler(
                name,
                self.registry,
                patterns=resource.watch.patterns or ["*"],  # type: ignore
                loop=self.loop,
            )
            self.handlers[name] = handler

            # Schedule directory watching for local paths only
            if hasattr(resource, "path"):
                path = upath.UPath(resource.path)  # type: ignore
                if is_watchable_path(path):
                    watch_path = str(path.parent)
                    if watch_path not in self._watched_paths:
                        self.observer.schedule(handler, watch_path, recursive=True)
                        self._watched_paths.add(watch_path)
                        logger.debug("Added watch for: %s -> %s", name, watch_path)
                else:
                    logger.debug(
                        "Skipping watch for non-local path: %s -> %s",
                        name,
                        path,
                    )
        except Exception:
            logger.exception("Failed to add watch for: %s", name)

    def remove_watch(self, name: str) -> None:
        """Remove a watch for a resource.

        Args:
            name: Resource name to unwatch
        """
        if _handler := self.handlers.pop(name, None):
            try:
                # Observer will be cleaned up on stop
                logger.debug("Removed watch for: %s", name)
            except Exception:
                logger.exception("Error removing watch for: %s", name)



"""Tests for MCP protocol implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest
import yaml

from llmling.config.models import Config, TextResource, ToolConfig
from llmling.prompts.models import Prompt, PromptMessage
from llmling.testing.testclient import MCPInProcSession


if TYPE_CHECKING:
    from collections.abc import AsyncGenerator
    from pathlib import Path


@pytest.fixture
def test_config() -> Config:
    """Create test configuration."""
    prompt = Prompt(
        name="test",
        description="test",
        messages=[PromptMessage(role="system", content="test")],
    )
    resource = TextResource(
        resource_type="text",
        content="Test content",
        description="Test resource",
    )
    tool_cfg = ToolConfig(
        import_path="llmling.testing.tools.example_tool",
        name="example",
        description="Test tool",
    )
    return Config(
        version="1.0",
        prompts={"test": prompt},
        resources={"test": resource},
        tools={"example": tool_cfg},
    )


@pytest.fixture
async def config_file(tmp_path: Path, test_config: Config) -> Path:
    """Create temporary config file."""
    config_path = tmp_path / "test_config.yml"
    content = test_config.model_dump(exclude_none=True)
    with config_path.open("w") as f:
        yaml.dump(content, f)
    return config_path


@pytest.fixture
async def configured_client(config_file: Path) -> AsyncGenerator[MCPInProcSession, None]:
    """Create client with test configuration."""
    client = MCPInProcSession(config_path=str(config_file))
    try:
        await client.start()
        response = await client.do_handshake()
        assert response["serverInfo"]["name"] == "llmling-server"
        yield client
    finally:
        await client.close()


@pytest.mark.asyncio
async def test_mcp_resource_operations(configured_client: MCPInProcSession):
    """Test MCP resource operations."""
    response = await configured_client.send_request("resources/list")
    assert "resources" in response
    resource_list = response["resources"]
    assert len(resource_list) >= 1
    assert resource_list[0]["name"] == "Test resource"


@pytest.mark.asyncio
async def test_mcp_tool_operations(configured_client: MCPInProcSession):
    """Test MCP tool operations."""
    # First verify tool exists
    tools = await configured_client.send_request("tools/list")
    tools_list = tools["tools"]
    assert len(tools_list) >= 1
    assert tools_list[0]["name"] == "example"

    # Now call it
    response = await configured_client.send_request(
        "tools/call",
        {
            "name": "example",
            "arguments": {"text": "test", "repeat": 1},
        },
    )
    assert "content" in response
    assert len(response["content"]) == 1
    assert response["content"][0]["text"] == "test"


@pytest.mark.asyncio
async def test_mcp_error_handling(configured_client: MCPInProcSession):
    """Test MCP error response format."""
    response = await configured_client.send_request("tools/call", {"name": "nonexistent"})
    assert "content" in response
    assert len(response["content"]) == 1
    assert "not found" in response["content"][0]["text"].lower()


@pytest.mark.asyncio
async def test_mcp_handshake(configured_client: MCPInProcSession):
    """Test MCP protocol handshake."""
    # Do another handshake to explicitly test it
    init_response = await configured_client.do_handshake()

    # Verify server info
    assert "serverInfo" in init_response
    assert init_response["serverInfo"]["name"] == "llmling-server"

    # Verify capabilities (they're on the root level, not in serverInfo)
    assert "capabilities" in init_response
    capabilities = init_response["capabilities"]

    # Verify specific capabilities
    assert "resources" in capabilities
    assert "prompts" in capabilities
    assert "tools" in capabilities


@pytest.mark.asyncio
async def test_mcp_prompt_operations(configured_client: MCPInProcSession):
    """Test MCP prompt operations."""
    # List prompts
    prompts = await configured_client.send_request("prompts/list")
    assert "prompts" in prompts
    assert isinstance(prompts["prompts"], list)

    # Get specific prompt
    # Add a test prompt if needed
    if prompts["prompts"]:
        test_prompt = prompts["prompts"][0]
        result = await configured_client.send_request(
            "prompts/get",
            {"name": test_prompt["name"], "arguments": None},
        )
        assert "messages" in result
        assert isinstance(result["messages"], list)


if __name__ == "__main__":
    pytest.main([__file__, "-v"])



"""Prompt management and composition."""

from __future__ import annotations

from typing import TYPE_CHECKING

from llmling.core.log import get_logger
from llmling.core.typedefs import Message
from llmling.prompts.models import MessageContext, PromptPriority, SystemPrompt


if TYPE_CHECKING:
    from llmling.tools.base import LLMCallableTool

logger = get_logger(__name__)


class PromptManager:
    """Manages prompt composition and organization."""

    def create_messages(
        self,
        context: MessageContext,
        *,
        tools: list[LLMCallableTool] | None = None,
    ) -> list[Message]:
        """Create messages from context.

        Args:
            context: Message construction context
            tools: Optional list of tools being used

        Returns:
            List of messages for LLM
        """
        messages: list[Message] = []

        # Collect all system prompts
        prompts = list(context.system_prompts)  # Copy original prompts

        # Add tool prompts if any tools provided
        if tools:
            prompts.extend(self._get_tool_prompts(tools))

        # Sort by priority and add to messages
        if prompts:
            msgs = self._create_system_messages(sorted(prompts, key=lambda p: p.priority))
            messages.extend(msgs)

        # Add user content
        if context.user_content or context.content_items:
            msg = Message(
                role="user",
                content=context.user_content,
                content_items=context.content_items,
            )
            messages.append(msg)

        return messages

    def _get_tool_prompts(
        self,
        tools: list[LLMCallableTool],
    ) -> list[SystemPrompt]:
        """Get system prompts from tools."""
        return [
            SystemPrompt(
                content=tool.system_prompt,
                source=f"tool:{tool.name}",
                priority=PromptPriority.TOOL,
                metadata={"tool_name": tool.name},
            )
            for tool in tools
            if hasattr(tool, "system_prompt")
        ]

    def _create_system_messages(
        self,
        prompts: list[SystemPrompt],
    ) -> list[Message]:
        """Create system messages from prompts."""
        return [
            Message(
                role="system",
                content=p.content,
                metadata={"source": p.source} if p.source else None,
            )
            for p in prompts
        ]



"""Collection of content processors for text transformation pipelines."""

from __future__ import annotations

from llmling.processors.base import (
    AsyncProcessor,
    BaseProcessor,
    ChainableProcessor,
    ProcessorConfig,
    ProcessorResult,
)
from llmling.processors.implementations.function import FunctionProcessor
from llmling.processors.implementations.template import TemplateProcessor
from llmling.processors.registry import ProcessorRegistry


__all__ = [
    "AsyncProcessor",
    "BaseProcessor",
    "ChainableProcessor",
    "FunctionProcessor",
    "ProcessorConfig",
    "ProcessorRegistry",
    "ProcessorResult",
    "TemplateProcessor",
]



"""Configuration validation utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING

import logfire

from llmling.core import exceptions
from llmling.core.log import get_logger


if TYPE_CHECKING:
    from llmling.config.models import Config


logger = get_logger(__name__)


class ConfigValidator:
    """Configuration validator."""

    def __init__(self, config: Config) -> None:
        """Initialize validator.

        Args:
            config: Configuration to validate
        """
        self.config = config

    def validate_all(self) -> list[str]:
        """Run all validations.

        Returns:
            List of validation warnings
        """
        warnings = []
        warnings.extend(self._validate_resources())
        warnings.extend(self._validate_processors())
        return warnings

    def _validate_resources(self) -> list[str]:
        """Validate resource configuration.

        Returns:
            List of validation warnings
        """
        warnings: list[str] = []

        warnings.extend(
            f"Resource {resource} in group {group} not found"
            for group, resources in self.config.resource_groups.items()
            for resource in resources
            if resource not in self.config.resources
        )

        warnings.extend(
            f"Processor {processor} in resource {name} not found"
            for name, resource in self.config.resources.items()
            for processor in resource.processors
            if processor.name not in self.config.context_processors
        )

        return warnings

    def _validate_processors(self) -> list[str]:
        """Validate processor configuration.

        Returns:
            List of validation warnings
        """
        warnings = []

        for name, processor in self.config.context_processors.items():
            if processor.type == "function" and not processor.import_path:
                warnings.append(
                    f"Processor {name} missing import_path for type 'function'",
                )
            elif processor.type == "template" and not processor.template:
                warnings.append(
                    f"Processor {name} missing template for type 'template'",
                )

        return warnings

    @logfire.instrument("Validating configs")
    def validate_or_raise(self) -> None:
        """Run all validations and raise on warnings.

        Raises:
            ConfigError: If any validation warnings are found
        """
        warnings = self.validate_all()
        if warnings:
            msg = "Configuration validation failed:\n" + "\n".join(warnings)
            raise exceptions.ConfigError(msg)



"""Registry for managing resources."""

from __future__ import annotations

from datetime import datetime
from typing import TYPE_CHECKING, Any

import logfire

from llmling.config.models import (
    CallableResource,
    CLIResource,
    ImageResource,
    PathResource,
    Resource,
    SourceResource,
    TextResource,
)
from llmling.core import exceptions
from llmling.core.baseregistry import BaseRegistry
from llmling.core.log import get_logger
from llmling.resources.watching import ResourceWatcher


if TYPE_CHECKING:
    from llmling.processors.registry import ProcessorRegistry
    from llmling.resources.loaders.registry import ResourceLoaderRegistry
    from llmling.resources.models import LoadedResource


logger = get_logger(__name__)


class ResourceRegistry(BaseRegistry[str, Resource]):
    """Registry for managing configured resources."""

    def __init__(
        self,
        loader_registry: ResourceLoaderRegistry,
        processor_registry: ProcessorRegistry,
    ) -> None:
        """Initialize registry with required dependencies."""
        super().__init__()
        self.loader_registry = loader_registry
        self.processor_registry = processor_registry
        # Cache by URI instead of name for consistency
        self._cache: dict[str, LoadedResource] = {}
        self._last_loaded: dict[str, datetime] = {}
        self.watcher = ResourceWatcher(self)

    def register(self, key: str, item: Resource | Any, replace: bool = False) -> None:
        """Register an item."""
        # Existing registration logic...
        super().register(key, item, replace)

        # Set up watching if needed
        if self._items[key].is_watched():
            self.watcher.add_watch(key, self._items[key])

    def __delitem__(self, key: str) -> None:
        """Remove item and its watch if present."""
        if key in self._items:
            # Remove watch first
            self.watcher.remove_watch(key)
        super().__delitem__(key)

    async def startup(self) -> None:
        """Initialize registry and start watcher."""
        await super().startup()
        # Start watcher
        await self.watcher.start()
        # Set up watches for existing resources
        for name, resource in self._items.items():
            if resource.is_watched():
                self.watcher.add_watch(name, resource)

    async def shutdown(self) -> None:
        """Cleanup registry and stop watcher."""
        # Stop watcher first
        await self.watcher.stop()
        # Then regular shutdown
        await super().shutdown()

    @property
    def _error_class(self) -> type[exceptions.ResourceError]:
        """Error class to use for this registry."""
        return exceptions.ResourceError

    def _validate_item(self, item: Any) -> Resource:
        """Validate and possibly transform items.

        Args:
            item: Item to validate

        Returns:
            Validated Resource

        Raises:
            ResourceError: If item is invalid
        """
        try:
            match item:
                # Match against concrete resource types
                case (
                    PathResource()
                    | TextResource()
                    | CLIResource()
                    | SourceResource()
                    | CallableResource()
                    | ImageResource()
                ):
                    return item

                case dict() if "resource_type" in item:
                    # Map resource_type to appropriate class
                    resource_classes: dict[str, type[Resource]] = {
                        "path": PathResource,
                        "text": TextResource,
                        "cli": CLIResource,
                        "source": SourceResource,
                        "callable": CallableResource,
                        "image": ImageResource,
                    }

                    resource_type = item["resource_type"]
                    if resource_type not in resource_classes:
                        msg = f"Unknown resource type: {resource_type}"
                        raise exceptions.ResourceError(msg)  # noqa: TRY301

                    # Validate using appropriate class
                    return resource_classes[resource_type].model_validate(item)

                case _:
                    msg = f"Invalid resource type: {type(item)}"
                    raise exceptions.ResourceError(msg)  # noqa: TRY301

        except Exception as exc:
            if isinstance(exc, exceptions.ResourceError):
                raise
            msg = f"Failed to validate resource: {exc}"
            raise exceptions.ResourceError(msg) from exc

    def get_uri(self, name: str) -> str:
        """Get URI for a resource by name."""
        resource = self[name]
        loader = self.loader_registry.get_loader(resource)
        return loader.create_uri(name=name)

    @logfire.instrument("Loading resource {name}")
    async def load(self, name: str, *, force_reload: bool = False) -> LoadedResource:
        """Load a resource by name."""
        try:
            resource = self[name]
            uri = self.get_uri(name)

            # Check cache unless force reload
            if not force_reload and uri in self._cache:
                return self._cache[uri]

            # Get loader and initialize with context
            loader = self.loader_registry.get_loader(resource)
            loader = loader.create(resource, name)  # Create with named context

            loaded = await loader.load(
                context=loader.context,  # Pass the context we created
                processor_registry=self.processor_registry,
            )

            # Ensure the URI is set correctly
            if loaded.metadata.uri != uri:
                logger.warning(
                    "Loader returned different URI than expected: %s != %s",
                    loaded.metadata.uri,
                    uri,
                )
                loaded.metadata.uri = uri

            # Update cache using URI
            self._cache[uri] = loaded
            self._last_loaded[uri] = datetime.now()
        except KeyError as exc:
            msg = f"Resource not found: {name}"
            raise exceptions.ResourceError(msg) from exc
        except Exception as exc:
            msg = f"Failed to load resource {name}: {exc}"
            raise exceptions.ResourceError(msg) from exc
        else:
            return loaded

    async def load_by_uri(self, uri: str) -> LoadedResource:
        """Load a resource by URI."""
        try:
            # Find loader that supports this URI
            loader = self.loader_registry.find_loader_for_uri(uri)

            # Get resource name from URI
            name = loader.get_name_from_uri(uri)

            # Load resource
            return await self.load(name)

        except Exception as exc:
            msg = f"Failed to load resource from URI {uri}"
            raise exceptions.ResourceError(msg) from exc

    def invalidate(self, name: str) -> None:
        """Invalidate cache for a resource."""
        try:
            uri = self.get_uri(name)
            self._cache.pop(uri, None)
            self._last_loaded.pop(uri, None)
        except Exception:
            logger.exception("Failed to invalidate resource: %s", name)

    def clear_cache(self) -> None:
        """Clear all cached resources."""
        self._cache.clear()
        self._last_loaded.clear()



from __future__ import annotations

from mcp.types import AnyUrl
import pytest

from llmling.server.conversions import from_mcp_uri, to_mcp_uri


@pytest.mark.parametrize(
    ("internal_uri", "mcp_uri"),
    [
        # File URIs
        ("file:///path/to/file.txt", "file://host/path/to/file.txt"),
        (
            "file:///C:/path/to/file.txt",
            "file://host/c/path/to/file.txt",
        ),  # Convert to /c/path format
        # HTTP URIs
        ("http://example.com", "http://example.com/"),
        ("https://example.com", "https://example.com/"),
        # Resource URIs
        ("text://content", "resource://host/content"),
        ("python://module.name", "resource://host/module.name"),
        ("cli://command", "resource://host/command"),
    ],
)
def test_uri_conversion_roundtrip(internal_uri: str, mcp_uri: str) -> None:
    """Test URI conversion in both directions."""
    # Internal -> MCP
    assert str(to_mcp_uri(internal_uri)) == mcp_uri
    # MCP -> Internal (for supported schemes)
    if not internal_uri.startswith(("text://", "python://", "cli://")):
        assert from_mcp_uri(mcp_uri) == internal_uri


@pytest.mark.parametrize(
    "uri",
    [
        "invalid://uri",
        "unknown://test",
        "://incomplete",
        "",  # Empty URI
    ],
)
def test_invalid_uri_conversion(uri: str) -> None:
    """Test handling of invalid URIs."""
    with pytest.raises(ValueError, match="URI"):
        to_mcp_uri(uri)
    with pytest.raises(ValueError, match="URI"):
        from_mcp_uri(uri)


def test_uri_type_handling() -> None:
    """Test handling of different URI input types."""
    # Test AnyUrl input
    mcp_uri = AnyUrl("file://host/test.txt")
    assert from_mcp_uri(str(mcp_uri)) == "file:///test.txt"

    # Test conversion to AnyUrl
    internal_uri = "file:///test.txt"
    assert isinstance(to_mcp_uri(internal_uri), AnyUrl)



"""Configuration loading utilities."""

from __future__ import annotations

from typing import TYPE_CHECKING

import logfire
from upath import UPath
import yamling

from llmling.config.models import Config
from llmling.config.validation import ConfigValidator
from llmling.core import exceptions
from llmling.core.log import get_logger


if TYPE_CHECKING:
    import os


logger = get_logger(__name__)


@logfire.instrument("Loading configuration from {path}")
def load_config(path: str | os.PathLike[str], *, validate: bool = True) -> Config:
    """Load and validate configuration from YAML file.

    Args:
        path: Path to configuration file
        validate: Whether to validate the configuration

    Returns:
        Loaded configuration

    Raises:
        ConfigError: If loading or validation fails
    """
    logger.debug("Loading configuration from %s", path)  # Use debug level

    try:
        content = yamling.load_yaml_file(path)
    except Exception as exc:
        msg = f"Failed to load YAML from {path!r}"
        raise exceptions.ConfigError(msg) from exc
    # Validate basic structure
    if not isinstance(content, dict):
        msg = "Configuration must be a dictionary"
        raise exceptions.ConfigError(msg)
    try:
        # Convert to model
        config = Config.model_validate(content)

        # Validate references if requested
        if validate:
            logger.debug("Validating configuration")  # Use debug level
            validator = ConfigValidator(config)
            validator.validate_or_raise()
    except Exception as exc:
        msg = f"Failed to load configuration from {path}"
        raise exceptions.ConfigError(msg) from exc
    else:
        msg = "Loaded configuration: version=%s, resources=%d"
        logger.info(msg, config.version, len(config.resources))
        return config


def save_config(
    config: Config,
    path: str | os.PathLike[str],
    *,
    validate: bool = True,
) -> None:
    """Save configuration to file.

    Args:
        config: Configuration to save
        path: Path to save to
        validate: Whether to validate before saving

    Raises:
        ConfigError: If validation or saving fails
    """
    logger.debug("Saving configuration to %s", path)

    try:
        if validate:
            validator = ConfigValidator(config)
            validator.validate_or_raise()

        # Convert to dictionary
        content = config.model_dump(exclude_none=True, by_alias=True)

        # Save to file
        string = yamling.dump_yaml(content)
        UPath(path).write_text(string)
        logger.info("Configuration saved successfully")

    except Exception as exc:
        msg = f"Failed to save configuration to {path}"
        raise exceptions.ConfigError(msg) from exc


if __name__ == "__main__":
    # Example usage
    import sys

    try:
        config_path = sys.argv[1] if len(sys.argv) > 1 else "config.yml"
        config = load_config(config_path)
        print(f"Loaded configuration version: {config.version}")
        print(f"Number of resources: {len(config.resources)}")
    except exceptions.ConfigError as e:
        print(f"Error loading config: {e}", file=sys.stderr)
        sys.exit(1)



"""Logging configuration for llmling."""

from __future__ import annotations

import logging
import queue
import sys
from typing import TYPE_CHECKING, Any

import platformdirs
from upath import UPath


if TYPE_CHECKING:
    from collections.abc import Sequence

    from mcp import types
    from mcp.server import Server


# Map Python logging levels to MCP logging levels
LEVEL_MAP: dict[int, types.LoggingLevel] = {
    logging.DEBUG: "debug",
    logging.INFO: "info",
    logging.WARNING: "warning",
    logging.ERROR: "error",
    logging.CRITICAL: "critical",
}

# Get platform-specific log directory
LOG_DIR = UPath(platformdirs.user_log_dir("llmling", "llmling"))
LOG_FILE = LOG_DIR / "llmling.log"

# Maximum log file size in bytes (10MB)
MAX_LOG_SIZE = 10 * 1024 * 1024
# Number of backup files to keep
BACKUP_COUNT = 5


def setup_logging(
    *,
    level: int | str = logging.INFO,
    handlers: Sequence[logging.Handler] | None = None,
    format_string: str | None = None,
    log_to_file: bool = True,
) -> None:
    """Configure logging for llmling.

    Args:
        level: The logging level for console output
        handlers: Optional sequence of handlers to add
        format_string: Optional custom format string
        log_to_file: Whether to log to file in addition to stdout
    """
    # Configure logfire first
    import logfire

    logfire.configure()

    logger = logging.getLogger("llmling")
    logger.setLevel(logging.DEBUG)  # Always set root logger to DEBUG

    if not format_string:
        format_string = "%(asctime)s - %(name)s - %(levelname)s - %(message)s"

    formatter = logging.Formatter(format_string)

    if not handlers:
        handlers = []
        # Add stdout handler with user-specified level
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setFormatter(formatter)
        stdout_handler.setLevel(level)
        handlers.append(stdout_handler)

        # Add file handler if requested (always DEBUG level)
        if log_to_file:
            try:
                # Create log directory if it doesn't exist
                LOG_DIR.mkdir(parents=True, exist_ok=True)

                # Use RotatingFileHandler for log rotation
                from logging.handlers import RotatingFileHandler

                file_handler = RotatingFileHandler(
                    LOG_FILE,
                    maxBytes=MAX_LOG_SIZE,
                    backupCount=BACKUP_COUNT,
                    encoding="utf-8",
                )
                file_handler.setFormatter(formatter)
                file_handler.setLevel(logging.DEBUG)  # Always DEBUG for file
                handlers.append(file_handler)
            except Exception as exc:  # noqa: BLE001
                print(
                    f"Failed to setup file logging at {LOG_FILE}: {exc}",
                    file=sys.stderr,
                )

    for handler in handlers:
        if not handler.formatter:
            handler.setFormatter(formatter)
        logger.addHandler(handler)

    # Log startup info with both handlers' levels
    logger.info("Logging initialized")
    if log_to_file:
        logger.debug(
            "Console logging level: %s, File logging level: DEBUG (%s)",
            logging.getLevelName(level),
            LOG_FILE,
        )


class MCPHandler(logging.Handler):
    """Handler that sends logs via MCP protocol."""

    def __init__(self, mcp_server: Server) -> None:
        """Initialize handler with MCP server instance."""
        super().__init__()
        self.server = mcp_server
        self.queue: queue.Queue[tuple[types.LoggingLevel, Any, str | None]] = (
            queue.Queue()
        )

    def emit(self, record: logging.LogRecord) -> None:
        """Queue log message for async sending."""
        try:
            # Try to get current session from server's request context
            try:
                _ = self.server.request_context  # Check if we have a context
            except LookupError:
                # No active session - fall back to stderr
                print(self.format(record), file=sys.stderr)
                return

            # Convert Python logging level to MCP level
            level = LEVEL_MAP.get(record.levelno, "info")

            # Format message
            message: Any = self.format(record)

            # Queue for async processing
            self.queue.put((level, message, record.name))

        except Exception:  # noqa: BLE001
            self.handleError(record)


def get_logger(name: str) -> logging.Logger:
    """Get a logger for the given name.

    Args:
        name: The name of the logger, will be prefixed with 'llmling.'

    Returns:
        A logger instance
    """
    return logging.getLogger(f"llmling.{name}")



"""Configuration file handling.

This module provides the ConfigFile class for managing YAML configuration files.
It handles loading, saving, and path resolution for LLMling configurations.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from upath import UPath
import yamling

from llmling.config.models import Config
from llmling.config.validation import ConfigValidator
from llmling.core import exceptions
from llmling.core.log import get_logger


if TYPE_CHECKING:
    import os


logger = get_logger(__name__)


class ConfigFile:
    """Manages a LLMling configuration file.

    This class handles loading and saving of configuration files, with built-in
    path resolution and validation. Configuration is loaded immediately upon
    instantiation and is always available through the `config` property.

    Args:
        path: Path to configuration file

    Raises:
        ConfigError: If loading or validation fails

    Example:
        >>> try:
        ...     config_file = ConfigFile("config.yml")
        ...     print(f"Loaded {len(config_file.config.resources)} resources")
        ... except exceptions.ConfigError as e:
        ...     print(f"Failed to load config: {e}")
    """

    def __init__(self, path: str | os.PathLike[str]) -> None:
        """Initialize and load configuration."""
        self.path = UPath(path).resolve()

        try:
            logger.debug("Loading configuration from %s", self.path)
            content = yamling.load_yaml_file(self.path)

            if not isinstance(content, dict):
                msg = "Configuration must be a dictionary"
                raise exceptions.ConfigError(msg)  # noqa: TRY301

            self._config = Config.model_validate(content)

            # Resolve paths relative to config location
            if hasattr(self._config, "resolve_paths"):
                self._config.resolve_paths(self.path.parent)

            logger.info(
                "Loaded configuration: version=%s, resources=%d",
                self._config.version,
                len(self._config.resources),
            )

        except Exception as exc:
            msg = f"Failed to load configuration from {self.path}"
            raise exceptions.ConfigError(msg) from exc

    def __repr__(self) -> str:
        """Return string representation of the config file."""
        return f"ConfigFile(path='{self.path}')"

    @property
    def config(self) -> Config:
        """The loaded configuration.

        Returns:
            Current configuration
        """
        return self._config

    def reload(self) -> None:
        """Reload configuration from file.

        This method re-reads the configuration file and updates the current
        configuration. All paths are re-resolved relative to the config file
        location.

        Raises:
            ConfigError: If reloading fails
        """
        self.__init__(self.path)  # type: ignore[misc]

    def save(self, *, validate: bool = True) -> None:
        """Save current configuration to file.

        Args:
            validate: Whether to validate before saving

        Raises:
            ConfigError: If saving or validation fails
        """
        try:
            if validate:
                validator = ConfigValidator(self._config)
                validator.validate_or_raise()

            content = self._config.model_dump(exclude_none=True, by_alias=True)
            string = yamling.dump_yaml(content)
            self.path.write_text(string)
            logger.info("Configuration saved to %s", self.path)

        except Exception as exc:
            msg = f"Failed to save configuration to {self.path}"
            raise exceptions.ConfigError(msg) from exc


def load_config(path: str | os.PathLike[str], *, validate: bool = True) -> Config:
    """Load and validate configuration from YAML file.

    Args:
        path: Path to configuration file
        validate: Whether to validate the configuration

    Returns:
        Loaded configuration

    Raises:
        ConfigError: If loading or validation fails
    """
    try:
        config_file = ConfigFile(path)

        if validate:
            validator = ConfigValidator(config_file.config)
            validator.validate_or_raise()
    except Exception as exc:
        msg = f"Failed to load config from {path}"
        raise exceptions.ConfigError(msg) from exc
    else:
        return config_file.config


if __name__ == "__main__":
    import sys

    try:
        config_path = sys.argv[1] if len(sys.argv) > 1 else "config.yml"
        config_file = ConfigFile(config_path)
        print(f"Loaded config file: {config_file}")
        print(f"Version: {config_file.config.version}")
        print(f"Resources: {len(config_file.config.resources)}")
    except exceptions.ConfigError as e:
        print(f"Error loading config: {e}", file=sys.stderr)
        sys.exit(1)



"""Template-based processor implementation."""

from __future__ import annotations

from typing import TYPE_CHECKING

import jinja2
import logfire

from llmling.core import exceptions
from llmling.core.log import get_logger
from llmling.processors.base import ChainableProcessor, ProcessorConfig, ProcessorResult


if TYPE_CHECKING:
    from llmling.resources.models import ProcessingContext


logger = get_logger(__name__)


class TemplateProcessor(ChainableProcessor):
    """Processor that applies a Jinja2 template."""

    def __init__(self, config: ProcessorConfig) -> None:
        """Initialize the processor."""
        super().__init__(config)
        self.template_config = config.get_template_config()
        self.template: jinja2.Template | None = None

    async def startup(self) -> None:
        """Compile template during startup."""
        try:
            loader = jinja2.BaseLoader()
            env = jinja2.Environment(loader=loader, autoescape=True, enable_async=True)
            self.template = env.from_string(self.config.template or "")
        except Exception as exc:
            msg = f"Failed to compile template: {exc}"
            raise exceptions.ProcessorError(msg) from exc

    @logfire.instrument("Rendering template")
    async def _process_impl(self, context: ProcessingContext) -> ProcessorResult:
        """Apply template to content."""
        if not self.template:
            msg = "Processor not initialized"
            raise exceptions.ProcessorError(msg)

        try:
            render_ctx = {"content": context.current_content, **context.kwargs}
            result = await self.template.render_async(**render_ctx)
            logger.debug("Template rendered: %s", result)
            keys = list(render_ctx.keys())
            meta = {"template_vars": keys, "template": self.config.template}
            orig = context.original_content
            return ProcessorResult(content=result, original_content=orig, metadata=meta)
        except Exception as exc:
            msg = f"Template rendering failed: {exc}"
            raise exceptions.ProcessorError(msg) from exc



"""Prompt-related models."""

from __future__ import annotations

from datetime import datetime  # noqa: TC003
from enum import Enum, IntEnum
from typing import Any, Literal

from pydantic import BaseModel, ConfigDict, Field, model_validator

from llmling.core.typedefs import MessageContent
from llmling.resources.models import LoadedResource  # noqa: TC001


MessageRole = Literal["system", "user", "assistant", "tool"]


class ArgumentType(str, Enum):
    """Types of prompt arguments that support completion."""

    TEXT = "text"
    FILE = "file"
    ENUM = "enum"
    RESOURCE = "resource"
    TOOL = "tool"


class PromptPriority(IntEnum):
    """Priority levels for system prompts."""

    TOOL = 100  # Tool-specific instructions
    SYSTEM = 200  # User-provided system prompts
    OVERRIDE = 300  # High-priority overrides


class SystemPrompt(BaseModel):
    """System prompt configuration."""

    content: str
    source: str = ""  # e.g., "tool:browser", "user", "config"
    priority: PromptPriority = PromptPriority.SYSTEM
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True)


class MessageContext(BaseModel):
    """Context for message construction."""

    system_prompts: list[SystemPrompt] = Field(default_factory=list)
    user_content: str = ""
    metadata: dict[str, Any] = Field(default_factory=dict)
    content_items: list[MessageContent] = Field(default_factory=list)

    model_config = ConfigDict(frozen=True)


class ExtendedPromptArgument(BaseModel):
    """Extended argument definition with completion support.

    This extends the base MCP PromptArgument with additional fields
    for completion and validation.
    """

    name: str
    description: str | None = None
    required: bool | None = None
    # Extended fields for completion support
    type: ArgumentType = ArgumentType.TEXT
    enum_values: list[str] | None = None  # For enum type
    file_patterns: list[str] | None = None  # For file type
    resource_types: list[str] | None = None  # For resource type
    tool_names: list[str] | None = None  # For tool type
    default: Any | None = None

    model_config = ConfigDict(extra="allow")

    @model_validator(mode="after")
    def validate_type_specific_fields(self) -> ExtendedPromptArgument:
        """Validate fields specific to argument types."""
        match self.type:
            case ArgumentType.ENUM:
                if not self.enum_values:
                    msg = "enum_values required for enum type"
                    raise ValueError(msg)
            case ArgumentType.FILE:
                if not self.file_patterns:
                    self.file_patterns = ["*"]  # Default to all files
            case ArgumentType.RESOURCE:
                if not self.resource_types:
                    self.resource_types = ["*"]  # Default to all resources
            case ArgumentType.TOOL:
                if not self.tool_names:
                    self.tool_names = ["*"]  # Default to all tools
        return self


class ResolvedContent(BaseModel):
    """Content with resolved resources."""

    original: MessageContent
    resolved: LoadedResource | None = None
    resolved_at: datetime | None = None

    model_config = ConfigDict(frozen=True)


class PromptMessage(BaseModel):
    """A message in a prompt."""

    role: MessageRole
    content: str | MessageContent | list[MessageContent] = ""
    resolved_content: list[ResolvedContent] | None = None

    model_config = ConfigDict(frozen=True)

    def needs_resolution(self) -> bool:
        """Check if message has unresolved resources."""
        contents = self.get_content_items()
        return any(item.type == "resource" for item in contents)

    def get_content_items(self) -> list[MessageContent]:
        """Get all content items."""
        match self.content:
            case str():
                return [MessageContent.text(self.content)]
            case MessageContent():
                return [self.content]
            case list():
                return self.content
            case _:
                return [MessageContent.text(str(self.content))]

    @model_validator(mode="before")
    @classmethod
    def ensure_content_items(cls, data: dict[str, Any]) -> dict[str, Any]:
        """Ensure backwards compatibility for content field."""
        if isinstance(data, dict):
            content = data.get("content", "")
            match content:
                case str():
                    # Convert string to text content
                    data["content"] = MessageContent.text(content)
                case MessageContent():
                    # Already correct format
                    pass
                case list():
                    # Ensure all items are MessageContent
                    data["content"] = [
                        item
                        if isinstance(item, MessageContent)
                        else MessageContent.text(str(item))
                        for item in content
                    ]
                case _:
                    # Convert anything else to string
                    data["content"] = MessageContent.text(str(content))
        return data

    def get_text_content(self) -> str:
        """Get text content for backwards compatibility."""
        match self.content:
            case str():
                return self.content
            case MessageContent() if self.content.type == "text":
                return self.content.content
            case list() if self.content:
                # Get first text content or first content
                text_items = [
                    item.content for item in self.content if item.type == "text"
                ]
                return text_items[0] if text_items else self.content[0].content
            case _:
                return ""


class Prompt(BaseModel):
    """Prompt template definition."""

    name: str
    description: str
    messages: list[PromptMessage]
    arguments: list[ExtendedPromptArgument] = Field(default_factory=list)
    metadata: dict[str, Any] = Field(default_factory=dict)

    model_config = ConfigDict(frozen=True)

    def validate_arguments(self, provided: dict[str, Any]) -> None:
        """Validate provided arguments against requirements."""
        required = {arg.name for arg in self.arguments if arg.required}
        missing = required - set(provided)
        if missing:
            msg = f"Missing required arguments: {', '.join(missing)}"
            raise ValueError(msg)


class PromptResult(BaseModel):
    """Result of rendering a prompt template."""

    messages: list[PromptMessage]
    metadata: dict[str, Any] = Field(default_factory=dict)
    resolved_at: datetime | None = None

    model_config = ConfigDict(frozen=True)



"""Models for file watching configuration."""

from __future__ import annotations

from pydantic import BaseModel, Field


class WatchConfig(BaseModel):
    """Watch configuration for resources."""

    enabled: bool = False
    patterns: list[str] | None = Field(
        default=None,
        description="List of pathspec patterns (.gitignore style)",
    )
    ignore_file: str | None = Field(
        default=None,
        description="Path to .gitignore-style file",
    )



"""Utilities for file watching."""

from __future__ import annotations

import asyncio
from functools import wraps
from threading import Lock
from typing import TYPE_CHECKING, Any

from upath import UPath

from llmling.core.log import get_logger


if TYPE_CHECKING:
    from collections.abc import Callable, Sequence
    import os


logger = get_logger(__name__)


def debounce(wait: float) -> Callable[[Callable[..., Any]], Callable[..., None]]:
    """Create a debounced version of a function."""

    def decorator(fn: Callable[..., Any]) -> Callable[..., None]:
        last_called = 0.0
        lock = Lock()

        @wraps(fn)
        def wrapped(*args: Any, **kwargs: Any) -> None:
            nonlocal last_called

            with lock:
                # Get main event loop
                try:
                    loop = asyncio.get_event_loop()
                    # Use call_soon_threadsafe to schedule from watchdog thread
                    loop.call_soon_threadsafe(fn, *args, **kwargs)
                except Exception:
                    logger.exception("Failed to schedule notification")

        return wrapped

    return decorator


def load_patterns(
    patterns: Sequence[str] | None = None,
    ignore_file: str | os.PathLike[str] | None = None,
) -> list[str]:
    """Load and combine watch patterns.

    Args:
        patterns: List of patterns from config
        ignore_file: Optional path to .gitignore style file

    Returns:
        Combined list of patterns
    """
    result: list[str] = []

    # Add configured patterns
    if patterns:
        result.extend(patterns)

    # Add patterns from file
    if ignore_file:
        try:
            path = UPath(ignore_file)
            if path.exists():
                # Filter empty lines and comments
                file_patterns = [
                    line.strip()
                    for line in path.read_text().splitlines()
                    if line.strip() and not line.startswith("#")
                ]
                result.extend(file_patterns)
        except Exception:
            logger.exception("Failed to load patterns from: %s", ignore_file)

    return result



"""Tests for configuration management."""

from __future__ import annotations

from typing import TYPE_CHECKING

import pytest

from llmling.config.manager import ConfigManager
from llmling.core import exceptions


if TYPE_CHECKING:
    from pathlib import Path


VERSION = "1.0"
EXPECTED_WARNING_COUNT = 2

CONFIG_YAML = f"""
version: "{VERSION}"
context_processors: {{}}
resources:
    test-context:
        type: text
        content: "Test content"
        description: "Test context"

resource_groups: {{}}
"""


@pytest.fixture
def config_file(tmp_path: Path) -> Path:
    """Create a test configuration file."""
    config_file = tmp_path / "config.yml"
    _ = config_file.write_text(CONFIG_YAML)
    return config_file


def test_load_config(config_file: Path):
    """Test loading configuration from file."""
    manager = ConfigManager.load(config_file)
    assert manager.config.version == VERSION


def test_load_invalid_config(tmp_path: Path):
    """Test loading invalid configuration."""
    invalid_file = tmp_path / "invalid.yml"
    _ = invalid_file.write_text("invalid: yaml: content")

    with pytest.raises(exceptions.ConfigError):
        _ = ConfigManager.load(invalid_file)


def test_save_config(tmp_path: Path, config_file: Path):
    """Test saving configuration."""
    manager = ConfigManager.load(config_file)

    save_path = tmp_path / "saved_config.yml"
    manager.save(save_path)

    # Load saved config and verify
    loaded = ConfigManager.load(save_path)
    assert loaded.config.model_dump() == manager.config.model_dump()


if __name__ == "__main__":
    _ = pytest.main(["-v", __file__])
